# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module defines the Mesh class, which holds the content (nodes,
elements, sets, ...) for a meshed geometry."""

import copy as _copy
import os as _os
import warnings as _warnings
from typing import Dict as _Dict
from typing import List as _List
from typing import Optional as _Optional

import numpy as _np
import pyvista as _pv
import quaternion as _quaternion
from numpy.typing import NDArray as _NDArray

from beamme.core.boundary_condition import (
    BoundaryConditionBase as _BoundaryConditionBase,
)
from beamme.core.boundary_condition import (
    BoundaryConditionContainer as _BoundaryConditionContainer,
)
from beamme.core.conf import bme as _bme
from beamme.core.coupling import coupling_factory as _coupling_factory
from beamme.core.element import Element as _Element
from beamme.core.element_beam import Beam as _Beam
from beamme.core.function import Function as _Function
from beamme.core.geometry_set import GeometryName as _GeometryName
from beamme.core.geometry_set import GeometrySetBase as _GeometrySetBase
from beamme.core.geometry_set import GeometrySetContainer as _GeometrySetContainer
from beamme.core.material import Material as _Material
from beamme.core.node import Node as _Node
from beamme.core.node import NodeCosserat as _NodeCosserat
from beamme.core.rotation import Rotation as _Rotation
from beamme.core.rotation import add_rotations as _add_rotations
from beamme.core.rotation import rotate_coordinates as _rotate_coordinates
from beamme.core.vtk_writer import VTKWriter as _VTKWriter
from beamme.geometric_search.find_close_points import (
    find_close_points as _find_close_points,
)
from beamme.geometric_search.find_close_points import (
    point_partners_to_partner_indices as _point_partners_to_partner_indices,
)
from beamme.utils.environment import is_testing as _is_testing
from beamme.utils.nodes import filter_nodes as _filter_nodes
from beamme.utils.nodes import find_close_nodes as _find_close_nodes
from beamme.utils.nodes import get_min_max_nodes as _get_min_max_nodes
from beamme.utils.nodes import get_nodal_coordinates as _get_nodal_coordinates
from beamme.utils.nodes import get_nodal_quaternions as _get_nodal_quaternions
from beamme.utils.nodes import get_nodes_by_function as _get_nodes_by_function


class Mesh:
    """A class that contains a full mesh, i.e. Nodes, Elements, Boundary
    Conditions, Sets, Couplings, Materials and Functions."""

    def __init__(self):
        """Initialize all empty containers."""

        self.nodes = []
        self.elements = []
        self.materials = []
        self.functions = []
        self.geometry_sets = _GeometrySetContainer()
        self.boundary_conditions = _BoundaryConditionContainer()

    @staticmethod
    def get_base_mesh_item_type(item):
        """Return the base mesh type of the given item.

        Amongst other things, we need this function so we can check if
        all items of a list are of the same "base" type.

        Args:
            item: The object we want to get the base type from.
        """

        for cls in (
            Mesh,
            _Function,
            _BoundaryConditionBase,
            _Material,
            _Node,
            _Element,
            _GeometrySetBase,
            _GeometryName,
        ):
            if isinstance(item, cls):
                return cls
        return type(item)

    def add(self, *args, **kwargs):
        """Add an item to this mesh, depending on its type.

        If an list is given each list element is added with this
        function. If multiple arguments are given, each one is
        individually added with this function. Keyword arguments are
        passed through to the adding function.
        """

        match len(args):
            case 0:
                raise ValueError("At least one argument is required!")
            case 1:
                add_item = args[0]
                base_type = self.get_base_mesh_item_type(add_item)

                base_type_to_method_map = {
                    Mesh: self.add_mesh,
                    _Function: self.add_function,
                    _BoundaryConditionBase: self.add_bc,
                    _Material: self.add_material,
                    _Node: self.add_node,
                    _Element: self.add_element,
                    _GeometrySetBase: self.add_geometry_set,
                    _GeometryName: self.add_geometry_name,
                    list: self.add_list,
                }
                if base_type in base_type_to_method_map:
                    base_type_to_method_map[base_type](add_item, **kwargs)
                else:
                    raise TypeError(
                        f'No Mesh.add case implemented for type: "{type(add_item)}" with base type "{base_type}"!'
                    )
            case _:
                for item in args:
                    self.add(item, **kwargs)

    def add_mesh(self, mesh):
        """Add the content of another mesh to this mesh."""

        # Add each item from mesh to self.
        self.add(mesh.nodes)
        self.add(mesh.elements)
        self.add(mesh.materials)
        self.add(mesh.functions)
        self.geometry_sets.extend(mesh.geometry_sets)
        self.boundary_conditions.extend(mesh.boundary_conditions)

    def add_bc(self, bc):
        """Add a boundary condition to this mesh."""
        bc_key = bc.bc_type
        geom_key = bc.geometry_set.geometry_type
        bc.geometry_set.check_replaced_nodes()
        self.boundary_conditions.append((bc_key, geom_key), bc)

    def add_function(self, function):
        """Add a function to this mesh item.

        Check that the function is only added once.
        """
        if function not in self.functions:
            self.functions.append(function)

    def add_material(self, material):
        """Add a material to this mesh item.

        Check that the material is only added once.
        """
        if material not in self.materials:
            self.materials.append(material)

    def add_node(self, node):
        """Add a node to this mesh."""
        if node in self.nodes:
            raise ValueError("The node is already in this mesh!")
        self.nodes.append(node)

    def add_element(self, element):
        """Add an element to this mesh."""
        if element in self.elements:
            raise ValueError("The element is already in this mesh!")
        self.elements.append(element)

    def add_geometry_set(self, geometry_set):
        """Add a geometry set to this mesh."""
        geometry_set.check_replaced_nodes()
        self.geometry_sets.append(geometry_set.geometry_type, geometry_set)

    def add_geometry_name(self, geometry_name):
        """Add a set of geometry sets to this mesh.

        Sort by the keys here to create a deterministic ordering,
        especially for testing purposes
        """
        keys = list(geometry_name.keys())
        keys.sort()
        for key in keys:
            self.add(geometry_name[key])

    def add_list(self, add_list: _List, **kwargs) -> None:
        """Add a list of items to this mesh.

        Args:
            add_list:
                List to be added to the mesh. This method checks that all
                base types of the list items are the same.

                In the special case of a node or element list, we add the whole list
                at once. This avoids a check for duplicate entries for every
                addition, which scales very badly. By doing it this way we only
                check the final list for duplicate entries which is much more
                performant.

                For all other types of items, we add each element individually
                via the Mesh.add method.
        """

        types = {self.get_base_mesh_item_type(item) for item in add_list}
        if len(types) > 1:
            raise TypeError(
                f"You can only add lists with the same type of element. Got {types}"
            )
        elif len(types) == 1:
            list_type = types.pop()

            def extend_internal_list(self_list: _List, new_list: _List) -> None:
                """Extend an internal list with the new list.

                It is checked that the final list does not have
                duplicate entries.
                """
                self_list.extend(new_list)
                if not len(set(self_list)) == len(self_list):
                    raise ValueError(
                        "The added list contains entries already existing in the Mesh"
                    )

            if list_type == _Node:
                extend_internal_list(self.nodes, add_list)
            elif list_type == _Element:
                extend_internal_list(self.elements, add_list)
            else:
                for item in add_list:
                    self.add(item, **kwargs)

    def replace_node(self, old_node, new_node):
        """Replace the first node with the second node."""

        # Check that the new node is in mesh.
        if new_node not in self.nodes:
            raise ValueError("The new node is not in the mesh!")

        for i, node in enumerate(self.nodes):
            if node == old_node:
                del self.nodes[i]
                break
        else:
            raise ValueError("The node that should be replaced is not in the mesh")

    def get_unique_geometry_sets(
        self,
        *,
        coupling_sets: bool = True,
        link_to_nodes: str = "no_link",
        geometry_set_start_indices: _Optional[_Dict] = None,
    ) -> _GeometrySetContainer:
        """Return a geometry set container that contains geometry sets
        explicitly added to the mesh, as well as sets for boundary conditions.

        The i_global values are set in the returned geometry sets.

        Args:
            coupling_sets:
                If this is true, also sets for couplings will be added.
            link_to_nodes:
                "no_link":
                    No link between the geometry set and the nodes is set
                "explicitly_contained_nodes":
                    A link will be set for all nodes that are explicitly part of the geometry set
                "all_nodes":
                    A link will be set for all nodes that are part of the geometry set, i.e., also
                    nodes connected to elements of an element set. This is mainly used for vtk
                    output so we can color the nodes which are part of element sets.
            geometry_set_start_indices:
                Dictionary where the keys are geometry type and the value is the starting
                index for those geometry sets. This can be used to define an offset in the
                i_global numbering of the geometry sets. The offsets default to 0.
        """

        is_link_nodes = not link_to_nodes == "no_link"
        if is_link_nodes:
            # First clear all links in existing nodes.
            for node in self.nodes:
                node.node_sets_link = []

        # Make a copy of the sets in this mesh.
        mesh_sets = self.geometry_sets.copy()

        # Add sets from boundary conditions.
        for (bc_key, geom_key), bc_list in self.boundary_conditions.items():
            for bc in bc_list:
                # Check if sets from couplings should be added.
                is_coupling = bc_key in (
                    _bme.bc.point_coupling,
                    bc_key == _bme.bc.point_coupling_penalty,
                )
                if (is_coupling and coupling_sets) or (not is_coupling):
                    # Only add set if it is not already in the container.
                    # For example if multiple Neumann boundary conditions
                    # are applied on the same node set.
                    if bc.geometry_set not in mesh_sets[geom_key]:
                        mesh_sets[geom_key].append(bc.geometry_set)

        for key in mesh_sets.keys():
            i_global_offset = 0
            if geometry_set_start_indices is not None:
                if key in geometry_set_start_indices:
                    i_global_offset = geometry_set_start_indices[key]
                else:
                    raise KeyError("Could not find {key} in geometry_set_start_indices")
            for i, geometry_set in enumerate(mesh_sets[key]):
                # Add global indices to the geometry set.
                geometry_set.i_global = i + i_global_offset
                if is_link_nodes:
                    geometry_set.link_to_nodes(link_to_nodes=link_to_nodes)

        return mesh_sets

    def set_node_links(self):
        """Create a link of all elements to the nodes connected to them.

        Also add a link to this mesh.
        """
        for element in self.elements:
            for node in element.nodes:
                node.element_link.append(element)
        for node in self.nodes:
            node.mesh = self

    def translate(self, vector):
        """Translate all beam nodes of this mesh.

        Args
        ----
        vector: _np.array, list
            3D vector that will be added to all nodes.
        """
        for node in self.nodes:
            node.coordinates += vector

    def rotate(
        self,
        rotation: _Rotation | _NDArray[_quaternion.quaternion],
        origin=None,
        only_rotate_triads: bool = False,
    ) -> None:
        """Rotate all beam nodes of the mesh with rotation.

        Args:
            rotation: The rotation(s) that will be applied to the nodes. If
            this is an array, it has to hold a quaternion for each node.
            origin (3D vector): If this is given, the mesh is rotated about
                this point. Defaults to (0, 0, 0).
            only_rotate_triads: If this is true, the nodal positions are not
                changed.
        """

        # Get array with all quaternions for the nodes.
        rot1 = _get_nodal_quaternions(self.nodes)

        # Apply the rotation to the rotation of all nodes.
        rot_new = _add_rotations(rotation, rot1)

        if not only_rotate_triads:
            # Get array with all positions for the nodes.
            pos = _get_nodal_coordinates(self.nodes)
            pos_new = _rotate_coordinates(pos, rotation, origin=origin)

        for i, node in enumerate(self.nodes):
            if isinstance(node, _NodeCosserat):
                node.rotation.q = rot_new[i, :]
            if not only_rotate_triads:
                node.coordinates = pos_new[i, :]

    def reflect(self, normal_vector, origin=None, flip_beams: bool = False) -> None:
        """Reflect all nodes of the mesh with respect to a plane defined by its
        normal_vector. Per default the plane goes through the origin, if not a
        point on the plane can be given with the parameter origin.

        For the reflection we assume that e1' and e2' are mirrored with respect
        to the original frame and e3' is in the opposite direction than the
        mirrored e3.

        With the defined mirroring strategy, the quaternion to be applied on
        the existing rotations can be calculated the following way:
            q[0] = e3 * n
            q[1,2,3] = e3 x n
        This constructs a rotation with the rotation axis on the plane, and
        normal to the vector e3. The rotation angle is twice the angle of e3
        to n.

        Args:
            normal_vector (3D vector): The normal vector of the reflection plane.
            origin (3D vector): The reflection plane goes through this point.
                Defaults to (0, 0, 0).
            flip_beams: When True, the beams are flipped, so that the direction
                along the beam is reversed.
        """

        # Normalize the normal vector.
        normal_vector = _np.asarray(normal_vector) / _np.linalg.norm(normal_vector)

        # Get array with all quaternions and positions for the nodes.
        pos = _get_nodal_coordinates(self.nodes)
        rot1 = _get_nodal_quaternions(self.nodes)

        # Check if origin has to be added.
        if origin is not None:
            pos -= origin

        # Get the reflection matrix A.
        A = _np.eye(3) - 2.0 * _np.outer(normal_vector, normal_vector)

        # Calculate the new positions.
        pos_new = _np.dot(pos, A)

        # Move back from the origin.
        if origin is not None:
            pos_new += origin

        # First get all e3 vectors of the nodes.
        e3 = _np.zeros_like(pos)
        e3[:, 0] = 2 * (rot1[:, 0] * rot1[:, 2] + rot1[:, 1] * rot1[:, 3])
        e3[:, 1] = 2 * (-1 * rot1[:, 0] * rot1[:, 1] + rot1[:, 2] * rot1[:, 3])
        e3[:, 2] = rot1[:, 0] ** 2 - rot1[:, 1] ** 2 - rot1[:, 2] ** 2 + rot1[:, 3] ** 2

        # Get the dot and cross product of e3 and the normal vector.
        rot2 = _np.zeros_like(rot1)
        rot2[:, 0] = _np.dot(e3, normal_vector)
        rot2[:, 1:] = _np.cross(e3, normal_vector)

        # Add to the existing rotations.
        rot_new = _add_rotations(rot2, rot1)

        if flip_beams:
            # To achieve the flip, the triads are rotated with the angle pi
            # around the e2 axis.
            rot_flip = _Rotation([0, 1, 0], _np.pi)
            rot_new = _add_rotations(rot_new, rot_flip)

        # For solid elements we need to adapt the connectivity to avoid negative Jacobians.
        # For beam elements this is optional.
        for element in self.elements:
            if isinstance(element, _Beam):
                if flip_beams:
                    element.flip()
            else:
                element.flip()

        # Set the new positions and rotations.
        for i, node in enumerate(self.nodes):
            node.coordinates = pos_new[i, :]
            if isinstance(node, _NodeCosserat):
                node.rotation.q = rot_new[i, :]

    def wrap_around_cylinder(
        self, radius: float | None = None, advanced_warning: bool = True
    ) -> None:
        """Wrap the geometry around a cylinder. The y-z plane gets morphed into
        the z-axis of symmetry. If all nodes are on the same y-z plane, the
        radius of the created cylinder is the x coordinate of that plane. If
        the nodes are not on the same y-z plane, the radius has to be given
        explicitly.

        Args:
            radius: If this value is given AND not all nodes are on the same y-z
                plane, then use this radius for the calculation of phi for all
                nodes. This might still lead to distorted elements!
        advanced_warning: If each element should be checked if it is either parallel
            to the y-z or x-z plane. This is computationally expensive, but in most
            cases (up to 100,000 elements) this check can be left activated.
        """

        pos = _get_nodal_coordinates(self.nodes)
        quaternions = _np.zeros([len(self.nodes), 4])

        # The x coordinate is the radius, the y coordinate the arc length.
        points_x = pos[:, 0].copy()

        # Check if all points are on the same y-z plane.
        if _np.abs(_np.min(points_x) - _np.max(points_x)) > _bme.eps_pos:
            # The points are not all on the y-z plane, get the reference
            # radius.
            if radius is not None:
                if advanced_warning:
                    # Here we check, if each element lays on a plane parallel
                    # to the y-z plane, or parallel to the x-z plane.
                    #
                    # To be exactly sure, we could check the rotations here,
                    # i.e. if they are also in plane.
                    element_warning = []
                    for i_element, element in enumerate(self.elements):
                        element_coordinates = _np.zeros([len(element.nodes), 3])
                        for i_node, node in enumerate(element.nodes):
                            element_coordinates[i_node, :] = node.coordinates
                        is_yz = (
                            _np.max(
                                _np.abs(
                                    element_coordinates[:, 0]
                                    - element_coordinates[0, 0]
                                )
                            )
                            < _bme.eps_pos
                        )
                        is_xz = (
                            _np.max(
                                _np.abs(
                                    element_coordinates[:, 1]
                                    - element_coordinates[0, 1]
                                )
                            )
                            < _bme.eps_pos
                        )
                        if not (is_yz or is_xz):
                            element_warning.append(i_element)
                    if len(element_warning) != 0:
                        _warnings.warn(
                            "There are elements which are not "
                            "parallel to the y-z or x-y plane. This will lead "
                            "to distorted elements!"
                        )
                else:
                    _warnings.warn(
                        "The nodes are not on the same y-z plane. "
                        "This may lead to distorted elements!"
                    )
            else:
                raise ValueError(
                    "The nodes that should be wrapped around a "
                    "cylinder are not on the same y-z plane. This will give "
                    "unexpected results. Give a reference radius!"
                )
            radius_phi = radius
            radius_points = points_x
        elif radius is None or _np.abs(points_x[0] - radius) < _bme.eps_pos:
            radius_points = radius_phi = points_x[0]
        else:
            raise ValueError(
                (
                    "The points are all on the same y-z plane with "
                    "the x-coordinate {} but the given radius {} is different. "
                    "This does not make sense."
                ).format(points_x[0], radius)
            )

        # Get the angle for all nodes.
        phi = pos[:, 1] / radius_phi

        # The rotation is about the z-axis.
        quaternions[:, 0] = _np.cos(0.5 * phi)
        quaternions[:, 3] = _np.sin(0.5 * phi)

        # Set the new positions in the global array.
        pos[:, 0] = radius_points * _np.cos(phi)
        pos[:, 1] = radius_points * _np.sin(phi)

        # Rotate the mesh
        self.rotate(quaternions, only_rotate_triads=True)

        # Set the new position for the nodes.
        for i, node in enumerate(self.nodes):
            node.coordinates = pos[i, :]

    def couple_nodes(
        self,
        *,
        nodes=None,
        reuse_matching_nodes=False,
        coupling_type=_bme.bc.point_coupling,
        coupling_dof_type=_bme.coupling_dof.fix,
    ):
        """Search through nodes and connect all nodes with the same
        coordinates.

        Args:
        ----
        nodes: [Node]
            List of nodes to couple. If None is given, all nodes of the mesh
            are coupled (except middle nodes).
        reuse_matching_nodes: bool
            If two nodes have the same position and rotation, the nodes are
            reduced to one node in the mesh. Be aware, that this might lead to
            issues if not all DOFs of the nodes should be coupled.
        coupling_type: bme.bc
            Type of point coupling.
        coupling_dof_type: str, bme.coupling_dof
            str: The string that will be used in the input file.
            bme.coupling_dof.fix: Fix all positional and rotational DOFs of the
                nodes together.
            bme.coupling_dof.joint: Fix all positional DOFs of the nodes
                together.
        """

        # Check that a coupling BC is given.
        if coupling_type not in (
            _bme.bc.point_coupling,
            _bme.bc.point_coupling_penalty,
        ):
            raise ValueError(
                "Only coupling conditions can be applied in 'couple_nodes'!"
            )

        # Get the nodes that should be checked for coupling. Middle nodes are
        # not checked, as coupling can only be applied to the boundary nodes.
        if nodes is None:
            node_list = self.nodes
        else:
            node_list = nodes
        node_list = _filter_nodes(node_list, middle_nodes=False)
        partner_nodes = _find_close_nodes(node_list)
        if len(partner_nodes) == 0:
            # If no partner nodes were found, end this function.
            return

        if reuse_matching_nodes:
            # Check if there are nodes with the same rotation. If there are the
            # nodes are reused, and no coupling is inserted.

            # Set the links to all nodes in the mesh. In this case we have to use
            # "all_nodes" since we also have to replace nodes that are in existing
            # GeometrySetNodes.
            self.unlink_nodes()
            self.get_unique_geometry_sets(link_to_nodes="explicitly_contained_nodes")
            self.set_node_links()

            # Go through partner nodes.
            for node_list in partner_nodes:
                # Get array with rotation vectors.
                rotation_vectors = _np.zeros([len(node_list), 3])
                for i, node in enumerate(node_list):
                    if isinstance(node, _NodeCosserat):
                        rotation_vectors[i, :] = node.rotation.get_rotation_vector()
                    else:
                        # For the case of nodes that belong to solid elements,
                        # we define the following default value:
                        rotation_vectors[i, :] = [4 * _np.pi, 0, 0]

                # Use find close points function to find nodes with the
                # same rotation.
                partners, n_partners = _find_close_points(
                    rotation_vectors, tol=_bme.eps_quaternion
                )

                # Check if nodes with the same rotations were found.
                if n_partners == 0:
                    self.add(
                        _coupling_factory(node_list, coupling_type, coupling_dof_type)
                    )
                else:
                    # There are nodes that need to be combined.
                    combining_nodes = []
                    coupling_nodes = []
                    found_partner_id = [None for _i in range(n_partners)]

                    # Add the nodes that need to be combined and add the nodes
                    # that will be coupled.
                    for i, partner in enumerate(partners):
                        if partner == -1:
                            # This node does not have a partner with the same
                            # rotation.
                            coupling_nodes.append(node_list[i])

                        elif found_partner_id[partner] is not None:
                            # This node has already a processed partner, add
                            # this one to the combining nodes.
                            combining_nodes[found_partner_id[partner]].append(
                                node_list[i]
                            )

                        else:
                            # This is the first node of a partner set that was
                            # found. This one will remain, the other ones will
                            # be replaced with this one.
                            new_index = len(combining_nodes)
                            found_partner_id[partner] = new_index
                            combining_nodes.append([node_list[i]])
                            coupling_nodes.append(node_list[i])

                    # Add the coupling nodes.
                    if len(coupling_nodes) > 1:
                        self.add(
                            _coupling_factory(
                                coupling_nodes, coupling_type, coupling_dof_type
                            )
                        )

                    # Replace the identical nodes.
                    for combine_list in combining_nodes:
                        master_node = combine_list[0]
                        for node in combine_list[1:]:
                            node.replace_with(master_node)

        else:
            # Connect close nodes with a coupling.
            for node_list in partner_nodes:
                self.add(_coupling_factory(node_list, coupling_type, coupling_dof_type))

    def unlink_nodes(self):
        """Delete the linked arrays and global indices in all nodes."""
        for node in self.nodes:
            node.unlink()

    def get_nodes_by_function(self, *args, **kwargs):
        """Return all nodes for which the function evaluates to true."""
        return _get_nodes_by_function(self.nodes, *args, **kwargs)

    def get_min_max_nodes(self, *args, **kwargs):
        """Return a geometry set with the max and min nodes in all
        directions."""
        return _get_min_max_nodes(self.nodes, *args, **kwargs)

    def check_overlapping_elements(self, raise_error=True):
        """Check if there are overlapping elements in the mesh.

        This is done by checking if all middle nodes of beam elements
        have unique coordinates in the mesh.
        """

        # Number of middle nodes.
        middle_nodes = [node for node in self.nodes if node.is_middle_node]

        # Only check if there are middle nodes.
        if len(middle_nodes) == 0:
            return

        # Get array with middle nodes.
        coordinates = _np.zeros([len(middle_nodes), 3])
        for i, node in enumerate(middle_nodes):
            coordinates[i, :] = node.coordinates

        # Check if there are double entries in the coordinates.
        has_partner, partner = _find_close_points(coordinates)
        partner_indices = _point_partners_to_partner_indices(has_partner, partner)
        if partner > 0:
            if raise_error:
                raise ValueError(
                    "There are multiple middle nodes with the "
                    "same coordinates. Per default this raises an error! "
                    "This check can be turned of with "
                    "bme.check_overlapping_elements=False"
                )
            else:
                _warnings.warn(
                    "There are multiple middle nodes with the same coordinates!"
                )

            # Add the partner index to the middle nodes.
            for i_partner, partners in enumerate(partner_indices):
                for i_node in partners:
                    middle_nodes[i_node].element_partner_index = i_partner

    def get_vtk_representation(
        self, *, overlapping_elements=True, coupling_sets=False, **kwargs
    ):
        """Return a vtk representation of the beams and solid in this mesh.

        Args
        ----
        overlapping_elements: bool
            I elements should be checked for overlapping. If they overlap, the
            output will mark them.
        coupling_sets: bool
            If coupling sets should also be displayed.
        """

        # Object to store VKT data (and write it to file)
        vtk_writer_beam = _VTKWriter()
        vtk_writer_solid = _VTKWriter()

        # Get the set numbers of the mesh
        mesh_sets = self.get_unique_geometry_sets(
            coupling_sets=coupling_sets, link_to_nodes="all_nodes"
        )

        # Set the global value for digits in the VTK output.
        # Get highest number of node_sets.
        max_sets = max(len(geometry_list) for geometry_list in mesh_sets.values())

        # Set the bme value.
        digits = len(str(max_sets))
        _bme.vtk_node_set_format = "{:0" + str(digits) + "}"

        if overlapping_elements:
            # Check for overlapping elements.
            self.check_overlapping_elements(raise_error=False)

        # Get representation of elements.
        for element in self.elements:
            element.get_vtk(vtk_writer_beam, vtk_writer_solid, **kwargs)

        # Finish and return the writers
        vtk_writer_beam.complete_data()
        vtk_writer_solid.complete_data()
        return vtk_writer_beam, vtk_writer_solid

    def write_vtk(
        self, output_name="beamme", output_directory="", binary=True, **kwargs
    ):
        """Write the contents of this mesh to VTK files.

        Args
        ----
        output_name: str
            Base name of the output file. There will be a {}_beam.vtu and
            {}_solid.vtu file.
        output_directory: path
            Directory where the output files will be written.
        binary: bool
            If the data should be written encoded in binary or in human readable text

        **kwargs
            For all of them look into:
                Mesh().get_vtk_representation
                Beam().get_vtk
                VolumeElement().get_vtk
        ----
        beam_centerline_visualization_segments: int
            Number of segments to be used for visualization of beam centerline between successive
            nodes. Default is 1, which means a straight line is drawn between the beam nodes. For
            Values greater than 1, a Hermite interpolation of the centerline is assumed for
            visualization purposes.
        """

        vtk_writer_beam, vtk_writer_solid = self.get_vtk_representation(**kwargs)

        # Write to file, only if there is at least one point in the writer.
        if vtk_writer_beam.points.GetNumberOfPoints() > 0:
            filepath = _os.path.join(output_directory, output_name + "_beam.vtu")
            vtk_writer_beam.write_vtk(filepath, binary=binary)
        if vtk_writer_solid.points.GetNumberOfPoints() > 0:
            filepath = _os.path.join(output_directory, output_name + "_solid.vtu")
            vtk_writer_solid.write_vtk(filepath, binary=binary)

    def display_pyvista(
        self,
        *,
        beam_nodes=True,
        beam_tube=True,
        beam_cross_section_directors=True,
        beam_radius_for_display=None,
        resolution=20,
        parallel_projection=False,
        **kwargs,
    ):
        """Display the mesh in pyvista.

        If this is called in a GitHub testing run, nothing will be shown, instead
        the _pv.plotter object will be returned.

        Args
        ----
        beam_nodes: bool
            If the beam nodes should be displayed. The start and end nodes of each
            beam will be shown in green, possible middle nodes inside the element
            are shown in cyan.
        beam_tube: bool
            If the beam should be rendered as a tube
        beam_cross_section_directors: bool
            If the cross section directors should be displayed (at each node)
        beam_radius_for_display: float
            If not all beams have an explicitly given radius (in the material
            definition) this value will be used to approximate the beams radius
            for visualization
        resolution: int
            Indicates how many triangulations will be performed to visualize arrows,
            tubes and spheres.
        parallel_projection: bool
            Flag to change camera view to parallel projection.

        **kwargs
            For all of them look into:
                Mesh().get_vtk_representation
                Beam().get_vtk
                VolumeElement().get_vtk
        ----
        beam_centerline_visualization_segments: int
            Number of segments to be used for visualization of beam centerline between successive
            nodes. Default is 1, which means a straight line is drawn between the beam nodes. For
            Values greater than 1, a Hermite interpolation of the centerline is assumed for
            visualization purposes.
        """

        plotter = _pv.Plotter()
        plotter.renderer.add_axes()

        if parallel_projection:
            plotter.enable_parallel_projection()

        vtk_writer_beam, vtk_writer_solid = self.get_vtk_representation(**kwargs)

        if vtk_writer_beam.points.GetNumberOfPoints() > 0:
            beam_grid = _pv.UnstructuredGrid(vtk_writer_beam.grid)

            # Check if all beams have a given cross-section radius, if not set the given input
            # value
            all_beams_have_cross_section_radius = (
                min(beam_grid.cell_data["cross_section_radius"]) > 0
            )
            if not all_beams_have_cross_section_radius:
                if beam_radius_for_display is None:
                    raise ValueError(
                        "Not all beams have a radius, you need to set "
                        "beam_radius_for_display to allow a display of the beams."
                    )
                beam_grid.cell_data["cross_section_radius"] = beam_radius_for_display

            # Grid with beam polyline
            beam_grid = beam_grid.cell_data_to_point_data()

            # Poly data for nodes
            finite_element_nodes = beam_grid.cast_to_poly_points().threshold(
                scalars="node_value", value=(0.4, 1.1)
            )

            # Plot the nodes
            node_radius_scaling_factor = 1.5
            if beam_nodes:
                sphere = _pv.Sphere(
                    radius=1.0,
                    theta_resolution=resolution,
                    phi_resolution=resolution,
                )
                nodes_glyph = finite_element_nodes.glyph(
                    geom=sphere,
                    scale="cross_section_radius",
                    factor=node_radius_scaling_factor,
                    orient=False,
                )
                plotter.add_mesh(
                    nodes_glyph.threshold(scalars="node_value", value=(0.9, 1.1)),
                    color="green",
                )
                middle_nodes = nodes_glyph.threshold(
                    scalars="node_value", value=(0.4, 0.6)
                )
                if len(middle_nodes.points) > 0:
                    plotter.add_mesh(middle_nodes, color="cyan")

            # Plot the beams
            beam_color = [0.5, 0.5, 0.5]
            if beam_tube:
                surface = beam_grid.extract_surface()
                if all_beams_have_cross_section_radius:
                    tube = surface.tube(
                        scalars="cross_section_radius",
                        absolute=True,
                        n_sides=resolution,
                    )
                else:
                    tube = surface.tube(
                        radius=beam_radius_for_display, n_sides=resolution
                    )
                plotter.add_mesh(tube, color=beam_color)
            else:
                plotter.add_mesh(beam_grid, color=beam_color, line_width=4)

            # Plot the directors of the beam cross-section
            director_radius_scaling_factor = 3.5
            if beam_cross_section_directors:
                arrow = _pv.Arrow(
                    tip_resolution=resolution, shaft_resolution=resolution
                )
                directors = [
                    finite_element_nodes.glyph(
                        geom=arrow,
                        orient=f"base_vector_{i + 1}",
                        scale="cross_section_radius",
                        factor=director_radius_scaling_factor,
                    )
                    for i in range(3)
                ]
                colors = ["white", "blue", "red"]
                for i, arrow in enumerate(directors):
                    plotter.add_mesh(arrow, color=colors[i])

        if vtk_writer_solid.points.GetNumberOfPoints() > 0:
            solid_grid = _pv.UnstructuredGrid(vtk_writer_solid.grid).clean()
            plotter.add_mesh(solid_grid, color="white", show_edges=True, opacity=0.5)

        if not _is_testing():
            plotter.show()
        else:
            return plotter

    def copy(self) -> "Mesh":
        """Return a deep copy of this mesh.

        The internal mesh data (nodes, elements, boundary conditions, and
        geometry sets) are deep-copied. Materials and functions are not
        deep-copied.

        **Important:** Some mesh creation functions return geometry set
        containers (e.g., node or element sets) that hold a reference to the
        nodes or elements of the mesh they were created with. When using
        ``mesh.copy()``, these externally returned sets remain linked to the
        original mesh and are therefore not transferred to the copied mesh.

        To copy both the mesh and the corresponding geometry sets correctly,
        deep-copy them together.

        Returns:
            A deep copy of the mesh.
        """
        return _copy.deepcopy(self)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file unit tests materials for 4C."""

import pytest

from beamme.four_c.material import (
    MaterialKirchhoff,
    MaterialReissner,
    MaterialReissnerElastoplastic,
    MaterialSolid,
    MaterialStVenantKirchhoff,
    get_all_contained_materials,
)


def test_beamme_four_c_material_reissner(assert_results_close):
    """Test Reissner material."""

    mat = MaterialReissner(
        radius=0.5,
        youngs_modulus=1234.56,
        nu=0.33,
        density=0.123,
        interaction_radius=2.4,
        shear_correction=17.15,
    )

    # set manually
    mat.i_global = 0

    mat_expected = {
        "MAT": 1,
        "MAT_BeamReissnerElastHyper": {
            "YOUNG": 1234.56,
            "POISSONRATIO": 0.33,
            "DENS": 0.123,
            "CROSSAREA": 0.7853981633974483,
            "SHEARCORR": 17.15,
            "MOMINPOL": 0.09817477042468103,
            "MOMIN2": 0.04908738521234052,
            "MOMIN3": 0.04908738521234052,
            "INTERACTIONRADIUS": 2.4,
        },
    }

    assert_results_close(mat.dump_to_list(), mat_expected)


def test_beamme_four_c_material_reissner_by_modes(assert_results_close):
    """Test Reissner material by modes with scaling factors."""

    mat = MaterialReissner(
        radius=0.5,
        youngs_modulus=1234.56,
        nu=0.33,
        density=0.123,
        interaction_radius=2.4,
        shear_correction=17.15,
        by_modes=True,
        scale_axial_rigidity=1.1,
        scale_shear_rigidity=1.2,
        scale_torsional_rigidity=1.3,
        scale_bending_rigidity=1.4,
    )

    # set manually
    mat.i_global = 0

    mat_expected = {
        "MAT": 1,
        "MAT_BeamReissnerElastHyper_ByModes": {
            "EA": 1066.5832722643493,
            "GA2": 7501.8057905674295,
            "GA3": 7501.8057905674295,
            "GI_T": 59.234375168474614,
            "EI2": 84.84185120284594,
            "EI3": 84.84185120284594,
            "RhoA": 0.09660397409788614,
            "MASSMOMINPOL": 0.012075496762235767,
            "MASSMOMIN2": 0.006037748381117884,
            "MASSMOMIN3": 0.006037748381117884,
            "INTERACTIONRADIUS": 2.4,
        },
    }

    assert_results_close(mat.dump_to_list(), mat_expected)


def test_beamme_four_c_material_reissner_elasto_plastic(assert_results_close):
    """Test the elasto plastic Reissner beam material."""

    kwargs = {
        "radius": 0.1,
        "nu": 1.0,
        "density": 1.0,
        "youngs_modulus": 1000,
        "interaction_radius": 2.0,
        "shear_correction": 5.0 / 6.0,
        "yield_moment": 2.3,
        "isohardening_modulus_moment": 4.5,
        "torsion_plasticity": False,
    }

    ref_dict = {
        "MAT": 69,
        "MAT_BeamReissnerElastPlastic": {
            "YOUNG": 1000,
            "POISSONRATIO": 1.0,
            "DENS": 1.0,
            "CROSSAREA": 0.031415926535897934,
            "SHEARCORR": 0.8333333333333334,
            "MOMINPOL": 0.00015707963267948968,
            "MOMIN2": 7.853981633974484e-05,
            "MOMIN3": 7.853981633974484e-05,
            "INTERACTIONRADIUS": 2.0,
            "YIELDM": 2.3,
            "ISOHARDM": 4.5,
            "TORSIONPLAST": False,
        },
    }

    mat = MaterialReissnerElastoplastic(**kwargs)
    mat.i_global = 68

    assert_results_close(mat.dump_to_list(), ref_dict)

    ref_dict["MAT_BeamReissnerElastPlastic"]["TORSIONPLAST"] = True
    kwargs["torsion_plasticity"] = True
    mat = MaterialReissnerElastoplastic(**kwargs)
    mat.i_global = 68
    assert_results_close(mat.dump_to_list(), ref_dict)


def test_beamme_four_c_material_kirchhoff_material(assert_results_close):
    """Test the Kirchhoff Love beam material."""

    def set_stiff(material):
        """Set the material properties for the beam material."""
        material.area = 2.0
        material.mom2 = 3.0
        material.mom3 = 4.0
        material.polar = 5.0

    material = MaterialKirchhoff(
        youngs_modulus=1000, radius=1.0, nu=1.0, density=1.0, is_fad=True
    )
    material.i_global = 26
    set_stiff(material)
    assert_results_close(
        material.dump_to_list(),
        {
            "MAT": 27,
            "MAT_BeamKirchhoffElastHyper": {
                "YOUNG": 1000,
                "SHEARMOD": 250.0,
                "DENS": 1.0,
                "CROSSAREA": 2.0,
                "MOMINPOL": 5.0,
                "MOMIN2": 3.0,
                "MOMIN3": 4.0,
                "FAD": True,
            },
        },
    )

    material = MaterialKirchhoff(
        youngs_modulus=1000, radius=1.0, nu=1.0, density=1.0, is_fad=False
    )
    material.i_global = 26
    set_stiff(material)
    assert_results_close(
        material.dump_to_list(),
        {
            "MAT": 27,
            "MAT_BeamKirchhoffElastHyper": {
                "YOUNG": 1000,
                "SHEARMOD": 250.0,
                "DENS": 1.0,
                "CROSSAREA": 2.0,
                "MOMINPOL": 5.0,
                "MOMIN2": 3.0,
                "MOMIN3": 4.0,
                "FAD": False,
            },
        },
    )

    material = MaterialKirchhoff(
        youngs_modulus=1000, radius=1.0, nu=1.0, density=1.0, interaction_radius=1.1
    )
    material.i_global = 26
    set_stiff(material)
    assert_results_close(
        material.dump_to_list(),
        {
            "MAT": 27,
            "MAT_BeamKirchhoffElastHyper": {
                "YOUNG": 1000,
                "SHEARMOD": 250.0,
                "DENS": 1.0,
                "CROSSAREA": 2.0,
                "MOMINPOL": 5.0,
                "MOMIN2": 3.0,
                "MOMIN3": 4.0,
                "FAD": False,
                "INTERACTIONRADIUS": 1.1,
            },
        },
    )


def test_beamme_four_c_material_stvenantkirchhoff_solid(assert_results_close):
    """Test that the solid with St_Venant Kirchhoff material."""

    material = MaterialStVenantKirchhoff(youngs_modulus=157, nu=0.17, density=6.1e-7)
    material.i_global = 3
    assert_results_close(
        material.dump_to_list(),
        {
            "MAT": 4,
            "MAT_Struct_StVenantKirchhoff": {
                "YOUNG": 157,
                "NUE": 0.17,
                "DENS": 6.1e-07,
            },
        },
    )


def test_beamme_four_c_material_sub_materials():
    """Test that sub-materials are correctly returned from the material."""

    material_1_1 = MaterialSolid(material_string="mat_1_1")
    material_1_2 = MaterialSolid(material_string="mat_1_2")
    material_1 = MaterialSolid(
        material_string="mat_1", data={"MATIDS": [material_1_1, material_1_2]}
    )

    material_2 = MaterialSolid(material_string="mat_2")

    material_3_1 = MaterialSolid(material_string="mat_3_1")
    material_3_2_1 = MaterialSolid(material_string="mat_3_2_1")
    material_3_2 = MaterialSolid(
        material_string="mat_3_2", data={"MATIDS": [material_3_2_1]}
    )
    material_3_3 = MaterialSolid(material_string="mat_3_3")
    material_3 = MaterialSolid(
        material_string="mat_3",
        data={"MATIDS": [material_3_1, material_3_2, material_3_3]},
    )

    material = MaterialSolid(
        material_string="mat", data={"MATIDS": [material_1, material_2, material_3]}
    )
    sub_materials = get_all_contained_materials(material)

    sub_materials_reference = [
        material,
        material_1,
        material_1_1,
        material_1_2,
        material_2,
        material_3,
        material_3_1,
        material_3_2,
        material_3_2_1,
        material_3_3,
    ]
    assert len(sub_materials_reference) == len(sub_materials)
    for mat_reference, mat_test in zip(sub_materials_reference, sub_materials):
        assert mat_reference is mat_test


def test_beamme_four_c_material_sub_materials_circular_loop():
    """Test that sub-materials containing circular loops are detected."""

    material_3 = MaterialSolid(material_string="mat_3", data={"MATIDS": [None]})
    material_2 = MaterialSolid(material_string="mat_2", data={"MATIDS": [material_3]})
    material_1 = MaterialSolid(material_string="mat_2", data={"MATIDS": [material_2]})

    # Set the circular reference
    material_3.data["MATIDS"][0] = material_1

    # Check that the circular reference is detected
    with pytest.raises(ValueError, match="Circular material reference detected!"):
        get_all_contained_materials(material_1)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Performance testing framework infrastructure."""

import time
import warnings
from dataclasses import dataclass
from typing import Callable, Optional

import pytest

from beamme.core.mesh import Mesh
from beamme.four_c.input_file import InputFile

PERFORMANCE_LOG = {}


@pytest.fixture(scope="module")
def cache_data():
    """Fixture to cache data across tests."""

    @dataclass
    class PyTestCache:
        """Cache for pytest data containing mesh and input file."""

        mesh: Optional[Mesh] = None
        input_file: Optional[InputFile] = None

    return PyTestCache()


@pytest.fixture(scope="function")
def evaluate_execution_time() -> Callable:
    """Return function to evaluate execution time.

    Necessary to enable the function call through pytest fixtures.

    Returns:
        Function to evaluate execution time.
    """

    def _evaluate_execution_time(
        name: str,
        function: Callable,
        *,
        args: tuple = (),
        kwargs: dict = {},
        expected_time: float,
    ):
        """Evaluate the execution time of a function and log it.

        Args:
            name: Name of the test for logging.
            function: The function to be executed.
            args: Positional arguments for the function.
            kwargs: Keyword arguments for the function.
            expected_time: Expected execution time in seconds.
        """

        start_time = time.time()
        result = function(*args, **kwargs)
        elapsed_time = time.time() - start_time

        PERFORMANCE_LOG[name] = {
            "execution_time": elapsed_time,
            "expected_time": expected_time,
        }

        if elapsed_time > expected_time:
            warnings.warn(
                f"Execution time for '{name}' test exceeded expected time: "
                f"{elapsed_time:.2f}s > {expected_time:.2f}s"
            )

        return result

    return _evaluate_execution_time


def pytest_terminal_summary(terminalreporter):
    """Print a summary of performance tests at the end of the pytest run."""

    if PERFORMANCE_LOG:
        terminalreporter.write_sep("=", "Performance Test Summary")

        for name, data in PERFORMANCE_LOG.items():
            elapsed = data["execution_time"]
            expected = data["expected_time"]

            if data["execution_time"] < data["expected_time"]:
                terminalreporter.write_line(
                    f"{name:<70} {elapsed:.3f}s < {expected:.3f}s",
                    bold=False,
                    green=True,
                )
            else:
                terminalreporter.write_line(
                    f"{name:<70} {elapsed:.3f}s > {expected:.3f}s",
                    bold=True,
                    red=True,
                )


def pytest_sessionfinish(session):
    """Exit with exit code 1 if any performance test failed.

    This is utilized to ensure that the Github Actions workflow fails if
    performance tests exceed their expected execution time.
    """

    if PERFORMANCE_LOG:
        for data in PERFORMANCE_LOG.values():
            if data["execution_time"] > data["expected_time"]:
                session.exitstatus = 1
                break



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file has functions to create a stent according to Auricchio 2012."""

import numpy as _np

from beamme.core.geometry_set import GeometryName as _GeometryName
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.mesh import Mesh as _Mesh
from beamme.core.rotation import Rotation as _Rotation
from beamme.mesh_creation_functions.beam_arc import (
    create_beam_mesh_arc_segment_via_rotation as _create_beam_mesh_arc_segment_via_rotation,
)
from beamme.mesh_creation_functions.beam_line import (
    create_beam_mesh_line as _create_beam_mesh_line,
)
from beamme.utils.nodes import get_min_max_nodes as _get_min_max_nodes


def create_stent_cell(
    beam_class,
    material,
    width,
    height,
    fac_bottom=0.6,
    fac_neck=0.55,
    fac_radius=0.36,
    alpha=0.46 * _np.pi,
    S1=True,
    S2=True,
    S3=True,
    n_el=1,
):
    """Create a cell of the stent. This cell is on the x-y plane.

    Args
    ----
    beam_class: Beam
        Class that will be used to create the beam elements.
    material: Material
        Material for the beam.
    width: float
        Width of the total cell.
    height: float
        Height of the total cell.
    fac_bottom: the ratio of the bottom's width to the cell's width
    fac_neck: the ratio of the neck's width to the cell's width
    fac_radius: the ratio of the S1's radius to the cell's width
    alpha: radiant
        The angle between the lines and horizontal line
    n_el: int
        Number of elements per beam line.
    S1, S2, S3: bool
        This check weather the curve S1, S2 or S3 will be created.
        If the cell is on bottom of the stent flat S1 and S2 won't
        be created. If the cell is on top of the flat S1 and S3
        won't be created

    ( these variables are described in a file )

    Return
    ----
    mesh: Mesh
        A mesh with this structure
    """

    mesh = _Mesh()

    def add_line(pointa, pointb, n_el_line):
        """Shortcut to add line."""
        return _create_beam_mesh_line(
            mesh, beam_class, material, pointa, pointb, n_el=n_el_line
        )

    def add_segment(center, axis_rotation, radius, angle, n_el_segment):
        """Shortcut to add arc segment."""
        return _create_beam_mesh_arc_segment_via_rotation(
            mesh,
            beam_class,
            material,
            center,
            axis_rotation,
            radius,
            angle,
            n_el=n_el_segment,
        )

    neck_width = width * fac_neck
    bottom_width = width * fac_bottom
    top_width = 2 * neck_width - bottom_width
    radius = width * fac_radius

    if S1:
        neck_point = _np.array([-neck_width, height * 0.5, 0])
        d = (height * 0.5 / _np.tan(alpha) + bottom_width - neck_width) / _np.sin(alpha)
        CM = _np.array([-_np.sin(alpha), -_np.cos(alpha), 0]) * (d - radius)
        MO = _np.array([_np.cos(alpha), -_np.sin(alpha), 0]) * _np.sqrt(
            radius**2 - (d - radius) ** 2
        )
        S1_angle = _np.pi / 2 + _np.arcsin((d - radius) / radius)
        S1_center1 = CM + MO + neck_point
        S1_axis_rotation1 = _Rotation([0, 0, 1], 2 * _np.pi - S1_angle - alpha)
        add_segment(S1_center1, S1_axis_rotation1, radius, S1_angle, n_el)
        add_line([-bottom_width, 0, 0], mesh.nodes[-1].coordinates, 2 * n_el)

        S1_center2 = 2 * neck_point - S1_center1
        S1_axis_rotation2 = _Rotation([0, 0, 1], _np.pi - alpha - S1_angle)

        add_segment(S1_center2, S1_axis_rotation2, radius, S1_angle, n_el)
        add_line(
            mesh.nodes[-1].coordinates, 2 * neck_point - [-bottom_width, 0, 0], 2 * n_el
        )

    if S3:
        S3_radius = (width - bottom_width + height * 0.5 / _np.tan(alpha)) * _np.tan(
            alpha / 2
        )
        S3_center = [-width, height * 0.5, 0] + S3_radius * _np.array([0, 1, 0])
        S3_axis_rotation = _Rotation()
        S3_angle = _np.pi - alpha
        add_segment(S3_center, S3_axis_rotation, S3_radius, S3_angle, n_el)
        add_line(mesh.nodes[-1].coordinates, [-bottom_width, height, 0], 2 * n_el)

    if S2:
        S2_radius = (height * 0.5 / _np.tan(alpha) + top_width) * _np.tan(alpha * 0.5)
        S2_center = [0, height * 0.5, 0] - S2_radius * _np.array([0, 1, 0])
        S2_angle = _np.pi - alpha
        S2_axis_rotation = _Rotation([0, 0, 1], _np.pi)
        add_segment(S2_center, S2_axis_rotation, S2_radius, S2_angle, 2 * n_el)
        add_line(mesh.nodes[-1].coordinates, [-top_width, 0, 0], 2 * n_el)

    mesh.translate([width, -height / 2, 0])
    return mesh


def create_stent_column(
    beam_class, material, width, height, n_height, n_el=1, **kwargs
):
    """Create a column of completed cells. A completed cell consists of one
    cell, that is created with the create cell function and it's reflection.

    Args
    ----
    beam_class: Beam
        Class that will be used to create the beam elements.
    material: Material
        Material for the beam.
    width: float
        Width of the total cell.
    height: float
        Height of the total cell.
    n_height: int
        The number of cells in the column
    n_el: int
        Number of elements per beam line.
    ( these variables are described in a file )

    Return
    ----
    mesh: Mesh
        A mesh with this structure.
    """

    mesh_column = _Mesh()
    for i in range(n_height):
        S1 = True
        S2 = True
        S3 = True
        if i == 0:
            S1 = False
            S2 = False
        if i == 1:
            S2 = False
        if i == n_height - 1:
            S1 = False
            S3 = False

        if i == n_height - 2:
            S3 = False
        unit_cell = create_stent_cell(
            beam_class,
            material,
            width,
            height,
            S1=S1,
            S2=S2,
            S3=S3,
            n_el=n_el,
            **kwargs,
        )
        unit_cell.translate([0, i * height, 0])
        mesh_column.add(unit_cell)

    column_copy = mesh_column.copy()
    column_copy.reflect(normal_vector=[1, 0, 0], origin=[width, 0, 0])
    mesh_column.add(column_copy)

    return mesh_column


def create_beam_mesh_stent_flat(
    beam_class, material, width_flat, height_flat, n_height, n_column, n_el=1, **kwargs
):
    """Create a flat stent structure on the x-y plane.

    Args
    ----
    beam_class: Beam
        Class that will be used to create the beam elements.
    material: Material
        Material for the beam.
    width_flat: float
        The width of the flat structure.
    height_flat: float
        The height of the flat structure.
    n_height: int
        The number of cells in y direction.
    n_column: int
        The number of columns in x direction.
    n_el: int
        Number of elements per beam line.

    Return
    ----
    mesh: Mesh
        A mesh with this structure
    """

    mesh_flat = _Mesh()
    width = width_flat / n_column / 2
    height = height_flat / n_height
    column_mesh = create_stent_column(
        beam_class, material, width, height, n_height, n_el=n_el, **kwargs
    )
    for i in range(n_column):
        column_copy = column_mesh.copy()
        column_copy.translate([2 * width * i, 0, 0])
        mesh_flat.add(column_copy)

    for i in range(n_column // 2):
        for j in range(n_height - 1):
            _create_beam_mesh_line(
                mesh_flat,
                beam_class,
                material,
                [4 * i * width, j * height, 0],
                [4 * i * width, (j + 1) * height, 0],
                n_el=2 * n_el,
            )
        _create_beam_mesh_line(
            mesh_flat,
            beam_class,
            material,
            [(4 * i + 2) * width, 0, 0],
            [(4 * i + 2) * width, height, 0],
            n_el=2 * n_el,
        )
    return mesh_flat


def create_beam_mesh_stent(
    mesh,
    beam_class,
    material,
    length,
    diameter,
    n_axis,
    n_circumference,
    **kwargs,
):
    """Create a stent structure around cylinder, The cylinder axis will be the
    z-axis.

    Args
    ----
    mesh: Mesh
        Mesh that the stent will be added to.
    beam_class: Beam
        Class that will be used to create the beam elements.
    material: Material
        Material for the beam.
    length: float
        The length of this stent.
    diameter: float
        The diameter of the stent's cross section.
    n_axis: int
        Number of cells in axial-direction.
    n_circumference: int
        Number of cells around the diameter.
    ( these variables are described in a file )

    Return
    ----
    return_set: GeometryName
        Set with nodes on the top, bottom boundaries. Those
        sets only contains end nodes of lines, not the middle ones.
        The set 'all' contains all nodes.
    """

    # Only allow even number of columns.
    if n_circumference % 2 == 1:
        raise ValueError("has to be even even number!")

    # Set the Parameter for other functions
    height_flat = length
    width_flat = _np.pi * diameter
    n_height = n_axis
    n_column = n_circumference

    mesh_stent = create_beam_mesh_stent_flat(
        beam_class, material, width_flat, height_flat, n_height, n_column, **kwargs
    )
    mesh_stent.rotate(_Rotation([1, 0, 0], _np.pi / 2))
    mesh_stent.rotate(_Rotation([0, 0, 1], _np.pi / 2))
    mesh_stent.translate([diameter / 2, 0, 0])
    mesh_stent.wrap_around_cylinder()

    # List of nodes from the stent that are candidates for connections.
    stent_nodes = [node for node in mesh_stent.nodes if node.is_end_node]

    # Add connections for the nodes with same positions.
    mesh_stent.couple_nodes(nodes=stent_nodes)

    # Get min and max nodes of the honeycomb.
    min_max_nodes = _get_min_max_nodes(stent_nodes)

    # Return the geometry set.
    return_set = _GeometryName()
    return_set["top"] = min_max_nodes["z_max"]
    return_set["bottom"] = min_max_nodes["z_min"]
    return_set["all"] = _GeometrySet(mesh_stent.elements)

    mesh.add_mesh(mesh_stent)

    return return_set



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
This script can be used to compile the cython code:
> python setup.py build_ext --inplace
"""

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup

extensions = [
    Extension(
        "beamme.geometric_search.cython_lib",
        ["src/beamme/geometric_search/cython_lib.pyx"],
        include_dirs=[np.get_include()],
    )
]

setup(
    ext_modules=cythonize(
        extensions,
        build_dir="src/build/cython_generated_code",
        annotate=True,
    ),
)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file defines the base volume element."""

import numpy as _np
import vtk as _vtk

from beamme.core.element import Element as _Element
from beamme.core.vtk_writer import add_point_data_node_sets as _add_point_data_node_sets


class VolumeElement(_Element):
    """A base class for a volume element."""

    # This class variables stores the information about the element shape in
    # vtk. And the connectivity to the nodes.
    vtk_cell_type = None
    vtk_topology: list = []

    def __init__(self, nodes=None, data={}, **kwargs):
        super().__init__(nodes=nodes, **kwargs)
        self.data = data

    def get_vtk(self, vtk_writer_beam, vtk_writer_solid, **kwargs):
        """Add the representation of this element to the VTK writer as a
        quad."""

        # Check that the element has a valid vtk cell type.
        if self.vtk_cell_type is None:
            raise TypeError(f"vtk_cell_type for {type(self)} not set!")

        # Dictionary with cell data.
        cell_data = {}

        # Dictionary with point data.
        point_data = {}

        # Array with nodal coordinates.
        coordinates = _np.zeros([len(self.nodes), 3])
        for i, node in enumerate(self.nodes):
            coordinates[i, :] = node.coordinates

        # Add the node sets connected to this element.
        _add_point_data_node_sets(point_data, self.nodes)

        # Add cell to writer.
        indices = vtk_writer_solid.add_points(coordinates, point_data=point_data)
        vtk_writer_solid.add_cell(
            self.vtk_cell_type, indices[self.vtk_topology], cell_data=cell_data
        )


class VolumeWEDGE6(VolumeElement):
    """A WEDGE6 volume element."""

    vtk_cell_type = _vtk.vtkWedge
    vtk_topology = list(range(6))


class VolumeHEX8(VolumeElement):
    """A HEX8 volume element."""

    vtk_cell_type = _vtk.vtkHexahedron
    vtk_topology = list(range(8))


class VolumeTET4(VolumeElement):
    """A TET4 volume element."""

    vtk_cell_type = _vtk.vtkTetra
    vtk_topology = list(range(4))


class VolumeTET10(VolumeElement):
    """A TET10 volume element."""

    vtk_cell_type = _vtk.vtkQuadraticTetra
    vtk_topology = list(range(10))


class VolumeHEX20(VolumeElement):
    """A HEX20 volume element."""

    vtk_cell_type = _vtk.vtkQuadraticHexahedron
    vtk_topology = [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        16,
        17,
        18,
        19,
        12,
        13,
        14,
        15,
    ]


class VolumeHEX27(VolumeElement):
    """A HEX27 volume element."""

    vtk_cell_type = _vtk.vtkTriQuadraticHexahedron
    vtk_topology = [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        16,
        17,
        18,
        19,
        12,
        13,
        14,
        15,
        24,
        22,
        21,
        23,
        20,
        25,
        26,
    ]



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Unit tests for the generic NURBS creation functions."""

import pytest

from beamme.core.mesh import Mesh
from beamme.mesh_creation_functions.nurbs_generic import (
    add_geomdl_nurbs_to_mesh,
    create_geometry_sets,
)
from beamme.mesh_creation_functions.nurbs_geometries import (
    create_nurbs_brick,
    create_nurbs_flat_plate_2d,
)


@pytest.mark.parametrize(
    ("nurbs_patch", "reference_values"),
    [
        (
            create_nurbs_flat_plate_2d(1, 2, n_ele_u=1, n_ele_v=2),
            {
                "vertex_u_min_v_min": [0],
                "vertex_u_min_v_max": [9],
                "vertex_u_max_v_min": [2],
                "vertex_u_max_v_max": [11],
                "line_v_min": [0, 1, 2],
                "line_v_min_next": [3, 4, 5],
                "line_v_max_next": [6, 7, 8],
                "line_v_max": [9, 10, 11],
                "line_u_min": [0, 3, 6, 9],
                "line_u_min_next": [1, 4, 7, 10],
                "line_u_max_next": [1, 4, 7, 10],
                "line_u_max": [2, 5, 8, 11],
                "surf": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
            },
        ),
        (
            create_nurbs_brick(1, 2, 3, n_ele_u=1, n_ele_v=2, n_ele_w=3),
            {
                "vertex_u_min_v_min_w_min": [0],
                "vertex_u_min_v_min_w_max": [48],
                "vertex_u_min_v_max_w_min": [9],
                "vertex_u_min_v_max_w_max": [57],
                "vertex_u_max_v_min_w_min": [2],
                "vertex_u_max_v_min_w_max": [50],
                "vertex_u_max_v_max_w_min": [11],
                "vertex_u_max_v_max_w_max": [59],
                "line_v_min_w_min": [0, 1, 2],
                "line_v_min_w_max": [48, 49, 50],
                "line_v_max_w_min": [9, 10, 11],
                "line_v_max_w_max": [57, 58, 59],
                "line_u_min_w_min": [0, 3, 6, 9],
                "line_u_min_w_max": [48, 51, 54, 57],
                "line_u_max_w_min": [2, 5, 8, 11],
                "line_u_max_w_max": [50, 53, 56, 59],
                "line_u_min_v_min": [0, 12, 24, 36, 48],
                "line_u_min_v_max": [9, 21, 33, 45, 57],
                "line_u_max_v_min": [2, 14, 26, 38, 50],
                "line_u_max_v_max": [11, 23, 35, 47, 59],
                "surf_u_min": [
                    0,
                    12,
                    24,
                    36,
                    48,
                    3,
                    15,
                    27,
                    39,
                    51,
                    6,
                    18,
                    30,
                    42,
                    54,
                    9,
                    21,
                    33,
                    45,
                    57,
                ],
                "surf_u_max": [
                    2,
                    14,
                    26,
                    38,
                    50,
                    5,
                    17,
                    29,
                    41,
                    53,
                    8,
                    20,
                    32,
                    44,
                    56,
                    11,
                    23,
                    35,
                    47,
                    59,
                ],
                "surf_v_min": [0, 12, 24, 36, 48, 1, 13, 25, 37, 49, 2, 14, 26, 38, 50],
                "surf_v_max": [
                    9,
                    21,
                    33,
                    45,
                    57,
                    10,
                    22,
                    34,
                    46,
                    58,
                    11,
                    23,
                    35,
                    47,
                    59,
                ],
                "surf_w_min": [0, 3, 6, 9, 1, 4, 7, 10, 2, 5, 8, 11],
                "surf_w_max": [48, 51, 54, 57, 49, 52, 55, 58, 50, 53, 56, 59],
                "vol": [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    9,
                    10,
                    11,
                    12,
                    13,
                    14,
                    15,
                    16,
                    17,
                    18,
                    19,
                    20,
                    21,
                    22,
                    23,
                    24,
                    25,
                    26,
                    27,
                    28,
                    29,
                    30,
                    31,
                    32,
                    33,
                    34,
                    35,
                    36,
                    37,
                    38,
                    39,
                    40,
                    41,
                    42,
                    43,
                    44,
                    45,
                    46,
                    47,
                    48,
                    49,
                    50,
                    51,
                    52,
                    53,
                    54,
                    55,
                    56,
                    57,
                    58,
                    59,
                ],
            },
        ),
    ],
)
def test_beamme_mesh_creation_functions_nurbs_generic_sets(
    nurbs_patch, reference_values
):
    """Test that the add NURBS to mesh functionality returns the correct
    geometry sets."""

    # Add the nurbs to a mesh
    mesh = Mesh()
    add_geomdl_nurbs_to_mesh(mesh, nurbs_patch)
    nurbs_patch = mesh.elements[0]

    # Create the geometry sets for this patch
    patch_set = create_geometry_sets(nurbs_patch)

    for key, geometry_set in patch_set.items():
        set_nodes = geometry_set.get_all_nodes()
        assert len(set_nodes) == len(reference_values[key])
        for i_node, node_index in enumerate(reference_values[key]):
            assert set_nodes[i_node] is mesh.nodes[node_index]



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test the functionality to create 4C input files."""

import numpy as np

from beamme.core.boundary_condition import BoundaryCondition
from beamme.core.conf import bme
from beamme.core.function import Function
from beamme.core.geometry_set import GeometrySet
from beamme.core.mesh import Mesh
from beamme.four_c.beam_potential import BeamPotential
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.four_c.input_file import InputFile
from beamme.mesh_creation_functions.beam_helix import create_beam_mesh_helix
from beamme.utils.nodes import is_node_on_plane


def test_four_c_beam_potential_helix(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the correct creation of input files for simulations including beam
    to beam potential interactions."""

    mesh = Mesh()
    input_file = InputFile()

    # define function for line charge density
    fun = Function([{"COMPONENT": 0, "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "t"}])
    mesh.add(fun)

    # define the beam potential
    beampotential = BeamPotential(
        pot_law_prefactor=[-1.0e-3, 12.45e-8],
        pot_law_exponent=[6.0, 12.0],
        pot_law_line_charge_density=[1.0, 2.0],
        pot_law_line_charge_density_funcs=[fun, None],
    )

    # set headers for static case and beam potential
    input_file.add(
        beampotential.create_header(
            potential_type="volume",
            cutoff_radius=10.0,
            evaluation_strategy="single_length_specific_small_separations_simple",
            regularization_type="linear",
            regularization_separation=0.1,
            integration_segments=2,
            gauss_points=50,
            potential_reduction_length=15.0,
            automatic_differentiation=False,
            choice_master_slave="smaller_eleGID_is_slave",
            runtime_output_interval_steps=1,
            runtime_output_force=True,
            runtime_output_moment=True,
            runtime_output_uids=True,
            runtime_output_per_ele_pair=True,
            runtime_output_every_iteration=True,
        )
    )

    # create helix
    helix_set = create_beam_mesh_helix(
        mesh,
        Beam3rHerm2Line3,
        get_default_test_beam_material(material_type="reissner"),
        [0.0, 0.0, 1.0],
        [0.0, 0.0, 0.0],
        [2.0, 0.0, 0.0],
        helix_angle=np.pi / 4,
        height_helix=10,
        n_el=4,
    )

    # add potential charge conditions to input file
    mesh.add(
        beampotential.create_potential_charge_conditions(geometry_set=helix_set["line"])
    )

    # Add boundary condition to bottom node
    mesh.add(
        BoundaryCondition(
            GeometrySet(
                mesh.get_nodes_by_function(
                    is_node_on_plane,
                    normal=[0, 0, 1],
                    origin_distance=0.0,
                    tol=0.1,
                )
            ),
            {
                "NUMDOF": 9,
                "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )

    input_file.add(mesh)

    assert_results_close(get_corresponding_reference_file_path(), input_file)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module implements containers to manage boundary conditions and geometry
sets in one object."""


class ContainerBase(dict):
    """A base class for containers."""

    def append(self, key, item):
        """Append item to this container and check if the item is already in
        the list corresponding to key."""

        type_ok = False
        for item_type in self.item_types:
            if isinstance(item, item_type):
                type_ok = True
                break
        if not type_ok:
            raise TypeError(
                f"You tried to add an item of type {type(item)}, but only types derived "
                + f"from {self.item_types} can be added"
            )
        if key not in self.keys():
            self[key] = []
        else:
            if item in self[key]:
                raise ValueError("The item is already in this container!")
        self[key].append(item)

    def extend(self, container):
        """Add all items of another container to this container."""

        if not isinstance(container, self.__class__):
            raise TypeError(
                f"Only containers of type {self.__class__} can be merged here, you tried "
                + f"add {type(container)}"
            )
        for key, items in container.items():
            for item in items:
                self.append(key, item)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Helper functions to find, filter and interact with nodes."""

from typing import Union as _Union

import numpy as _np
from numpy.typing import NDArray as _NDArray

from beamme.core.conf import bme as _bme
from beamme.core.geometry_set import GeometryName as _GeometryName
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.geometry_set import GeometrySetBase as _GeometrySetBase
from beamme.core.node import Node as _Node
from beamme.core.node import NodeCosserat as _NodeCosserat
from beamme.geometric_search.find_close_points import (
    find_close_points as _find_close_points,
)
from beamme.geometric_search.find_close_points import (
    point_partners_to_partner_indices as _point_partners_to_partner_indices,
)


def find_close_nodes(nodes, **kwargs):
    """Find nodes in a point cloud that are within a certain tolerance of each
    other.

    Args
    ----
    nodes: list(Node)
        Nodes who are part of the point cloud.
    **kwargs:
        Arguments passed on to geometric_search.find_close_points

    Return
    ----
    partner_nodes: list(list(Node))
        A list of lists of nodes that are close to each other, i.e.,
        each element in the returned list contains nodes that are close
        to each other.
    """

    coords = _np.zeros([len(nodes), 3])
    for i, node in enumerate(nodes):
        coords[i, :] = node.coordinates
    partner_indices = _point_partners_to_partner_indices(
        *_find_close_points(coords, **kwargs)
    )
    return [[nodes[i] for i in partners] for partners in partner_indices]


def adjust_close_nodes(nodes: list[_Node], *, tol=_bme.eps_pos) -> None:
    """Adjust the coordinates of nodes that are within the given tolerance by
    setting all involved coordinates of the nodes to their common mean.

    Args:
        nodes: List of nodes whose coordinates need adjustment.
        tol: Distance tolerance used to detect partner nodes.
    """

    partner_nodes = find_close_nodes(nodes, tol=tol)
    for close_nodes in partner_nodes:
        average_coords = _np.mean([node.coordinates for node in close_nodes], axis=0)
        for node in close_nodes:
            node.coordinates = average_coords.copy()


def check_node_by_coordinate(node, axis, value, eps=_bme.eps_pos):
    """Check if the node is at a certain coordinate value.

    Args
    ----
    node: Node
        The node to be checked for its position.
    axis: int
        Coordinate axis to check.
        0 -> x, 1 -> y, 2 -> z
    value: float
        Value for the coordinate that the node should have.
    eps: float
        Tolerance to check for equality.
    """
    return _np.abs(node.coordinates[axis] - value) < eps


def get_min_max_coordinates(nodes):
    """Return an array with the minimal and maximal coordinates of the given
    nodes.

    Return
    ----
    min_max_coordinates:
        [min_x, min_y, min_z, max_x, max_y, max_z]
    """
    coordinates = _np.zeros([len(nodes), 3])
    for i, node in enumerate(nodes):
        coordinates[i, :] = node.coordinates
    min_max = _np.zeros(6)
    min_max[:3] = _np.min(coordinates, axis=0)
    min_max[3:] = _np.max(coordinates, axis=0)
    return min_max


def get_single_node(item: _Union[_Node, _GeometrySetBase]) -> _NodeCosserat:
    """Function to get a single node from the input item.

    Args:
        item: This can be a GeometrySet with exactly one node or a single node object.

    Returns:
        If a single node, or a Geometry set (point set) containing a single node
        is given, that node is returned, otherwise an error is raised.
    """
    if isinstance(item, _Node):
        node = item
    elif isinstance(item, _GeometrySetBase):
        # Check if there is only one node in the set
        nodes = item.get_points()
        if len(nodes) == 1:
            node = nodes[0]
        else:
            raise ValueError("GeometrySet does not have exactly one node!")
    else:
        raise TypeError(
            f'The given object can be node or GeometrySet got "{type(item)}"!'
        )

    if not isinstance(node, _NodeCosserat):
        raise TypeError("Expected a NodeCosserat object.")

    return node


def filter_nodes(nodes, *, middle_nodes=True):
    """Filter the list of the given nodes. Be aware that if no filters are
    enabled the original list will be returned.

    Args
    ----
    nodes: list(Nodes)
        If this list is given it will be returned as is.
    middle_nodes: bool
        If middle nodes should be returned or not.
    """

    if not middle_nodes:
        return [node for node in nodes if middle_nodes or not node.is_middle_node]
    else:
        return nodes


def get_nodal_coordinates(nodes):
    """Return an array with the coordinates of the given nodes.

    Args
    ----
    kwargs:
        Will be passed to self.get_global_nodes.

    Return
    ----
    pos: _np.array
        Numpy array with all the positions of the nodes.
    """
    coordinates = _np.zeros([len(nodes), 3])
    for i, node in enumerate(nodes):
        coordinates[i, :] = node.coordinates
    return coordinates


def get_nodal_quaternions(nodes: list[_Node]) -> _NDArray:
    """Return an array with the quaternions of the given nodes.

    Args:
        nodes: List of nodes where we want the quaternion array.
    Returns:
        A numpy array containing the quaternions (the length is the number of
        nodes and the dtype is a numpy quaternion). For nodes which don't
        contain a rotation, we set the dummy quaternion (2, 0, 0, 0).
    """
    quaternions = _np.zeros([len(nodes), 4])
    for i, node in enumerate(nodes):
        if isinstance(node, _NodeCosserat):
            quaternions[i, :] = node.rotation.get_quaternion()
        else:
            # For the case of nodes that belong to solid elements,
            # we define the following default value:
            quaternions[i, :] = [2.0, 0.0, 0.0, 0.0]
    return quaternions


def get_nodes_by_function(nodes, function, *args, middle_nodes=False, **kwargs):
    """Return all nodes for which the function evaluates to true.

    Args
    ----
    nodes: [Node]
        Nodes that should be filtered.
    function: function(node, *args, **kwargs)
        Nodes for which this function is true are returned.
    middle_nodes: bool
        If this is true, middle nodes of a beam are also returned.
    """
    node_list = filter_nodes(nodes, middle_nodes=middle_nodes)
    return [node for node in node_list if function(node, *args, **kwargs)]


def get_min_max_nodes(nodes, *, middle_nodes=False):
    """Return a geometry set with the max and min nodes in all directions.

    Args
    ----
    nodes: list(Nodes)
        If this one is given return an array with the coordinates of the
        nodes in list, otherwise of all nodes in the mesh.
    middle_nodes: bool
        If this is true, middle nodes of a beam are also returned.
    """

    node_list = filter_nodes(nodes, middle_nodes=middle_nodes)
    geometry = _GeometryName()

    pos = get_nodal_coordinates(node_list)
    for i, direction in enumerate(["x", "y", "z"]):
        # Check if there is more than one value in dimension.
        min_max = [_np.min(pos[:, i]), _np.max(pos[:, i])]
        if _np.abs(min_max[1] - min_max[0]) >= _bme.eps_pos:
            for j, text in enumerate(["min", "max"]):
                # get all nodes with the min / max coordinate
                min_max_nodes = []
                for index, value in enumerate(
                    _np.abs(pos[:, i] - min_max[j]) < _bme.eps_pos
                ):
                    if value:
                        min_max_nodes.append(node_list[index])
                geometry[f"{direction}_{text}"] = _GeometrySet(min_max_nodes)
    return geometry


def is_node_on_plane(
    node, *, normal=None, origin_distance=None, point_on_plane=None, tol=_bme.eps_pos
):
    """Query if a node lies on a plane defined by a point_on_plane or the
    origin distance.

    Args
    ----
    node:
        Check if this node coincides with the defined plane.
    normal: _np.array, list
        Normal vector of defined plane.
    origin_distance: float
        Distance between origin and defined plane. Mutually exclusive with
        point_on_plane.
    point_on_plane: _np.array, list
        Point on defined plane. Mutually exclusive with origin_distance.
    tol: float
        Tolerance of evaluation if point coincides with plane

    Return
    ----
    True if the point lies on the plane, False otherwise.
    """

    if origin_distance is None and point_on_plane is None:
        raise ValueError("Either provide origin_distance or point_on_plane!")
    elif origin_distance is not None and point_on_plane is not None:
        raise ValueError("Only provide origin_distance OR point_on_plane!")

    if origin_distance is not None:
        projection = _np.dot(node.coordinates, normal) / _np.linalg.norm(normal)
        distance = _np.abs(projection - origin_distance)
    elif point_on_plane is not None:
        distance = _np.abs(
            _np.dot(point_on_plane - node.coordinates, normal) / _np.linalg.norm(normal)
        )

    return distance < tol



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Unit tests for the run 4C module."""

from unittest import mock

import pytest

from beamme.four_c.run_four_c import clean_simulation_directory


@pytest.fixture
def create_simulation_dir(tmp_path):
    """Fixture to create a dummy simulation directory with a file and
    subdirectory."""
    simulation_dir = tmp_path / "existing_dir"
    simulation_dir.mkdir()
    simulation_sub_dir = simulation_dir / "sub"
    simulation_sub_dir.mkdir()
    test_file = simulation_dir / "test.txt"
    test_file.write_text("content")
    return simulation_dir, simulation_sub_dir, test_file


def test_beamme_four_c_run_four_c_clean_simulation_directory_create_directory(tmp_path):
    """Test that clean_simulation_directory creates a directory."""
    simulation_dir = tmp_path / "new_dir"
    assert not simulation_dir.exists()

    clean_simulation_directory(simulation_dir)
    assert simulation_dir.exists()


@pytest.mark.parametrize("prompt", (False, True))
def test_beamme_four_c_run_four_c_clean_simulation_directory_clear_existing_directory(
    prompt, create_simulation_dir
):
    """Test that clean_simulation_directory cleans an existing directory."""
    simulation_dir, _, _ = create_simulation_dir

    if prompt:
        with mock.patch("builtins.input", return_value="y"):
            clean_simulation_directory(simulation_dir, ask_before_clean=True)
    else:
        clean_simulation_directory(simulation_dir, ask_before_clean=False)

    assert simulation_dir.exists()
    assert not any(simulation_dir.iterdir())


def test_beamme_four_c_run_four_c_clean_simulation_directory_prompt_no(
    create_simulation_dir,
):
    """Test that clean_simulation_directory NO prompt works."""
    simulation_dir, simulation_sub_dir, test_file = create_simulation_dir

    with mock.patch("builtins.input", return_value="n"):
        with pytest.raises(ValueError, match="Directory is not deleted!"):
            clean_simulation_directory(simulation_dir, ask_before_clean=True)

    assert simulation_dir.exists()
    assert simulation_sub_dir.exists()
    assert test_file.exists()



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Unit tests for the NURBS utils functions."""

import numpy as np
import pytest

from beamme.core.rotation import Rotation
from beamme.mesh_creation_functions.nurbs_utils import (
    ensure_3d_splinepy_object,
    rotate_splinepy,
    translate_splinepy,
)


class MockSpline:
    """Minimal mock class to mimic a splinepy object with control points.

    By using this mock, we can test that the translate and rotate
    functions only modify the control points.
    """

    def __init__(self, control_points):
        self.control_points = np.array(control_points)


def test_beamme_mesh_creation_functions_nurbs_utils_translate_splinepy_2d(
    assert_results_close,
):
    """Test translation of a 2D splinepy object."""
    spline = MockSpline([[0.0, 0.0], [1.0, 1.0]])
    translation = [2.0, -1.0]
    translate_splinepy(spline, translation)
    expected = np.array([[2.0, -1.0], [3.0, 0.0]])
    assert_results_close(spline.control_points, expected)


def test_beamme_mesh_creation_functions_nurbs_utils_translate_splinepy_3d(
    assert_results_close,
):
    """Test translation of a 3D splinepy object."""
    spline = MockSpline([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
    translation = [1.0, 2.0, 3.0]
    translate_splinepy(spline, translation)
    expected = np.array([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]])
    assert_results_close(spline.control_points, expected)


def test_beamme_mesh_creation_functions_nurbs_utils_translate_invalid_dimension():
    """Test that translation with a vector of incorrect dimension raises
    error."""
    spline = MockSpline([[0.0, 0.0]])
    with pytest.raises(ValueError, match="Dimensions of translation"):
        translate_splinepy(spline, [1.0, 2.0, 3.0])


def test_beamme_mesh_creation_functions_nurbs_utils_rotate_splinepy_2d(
    assert_results_close,
):
    """Test rotation of a 2D splinepy object."""
    points = [[1.0, 0.0], [0.0, 1.0]]
    spline = MockSpline(points)
    rotation = Rotation([0, 0, 1], np.pi / 2)
    rotate_splinepy(spline, rotation)
    expected = np.array([[0.0, 1.0], [-1.0, 0.0]])
    assert_results_close(spline.control_points, expected)


def test_beamme_mesh_creation_functions_nurbs_utils_rotate_splinepy_invalid_2d_rotation():
    """Test that a 2D splinepy object can only be rotated by a rotation around
    the z-axis."""
    spline = MockSpline([[1.0, 0.0]])
    rotation = Rotation([1, 1, 0], np.pi / 3)
    with pytest.raises(ValueError, match="Rotation vector must be in"):
        rotate_splinepy(spline, rotation)


@pytest.mark.parametrize("origin", [None, [1.0, 2.0, 3.0]])
def test_beamme_mesh_creation_functions_nurbs_utils_rotate_splinepy_3d_origin(
    origin, assert_results_close
):
    """Test rotation of a 3D splinepy object with and without given origin
    argument."""
    points = np.array([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]])
    spline = MockSpline(points)
    rotation = Rotation([1.0, 2.0, 3.0], np.pi / 3.0)
    rotate_splinepy(spline, rotation, origin=origin)
    origin_for_reference = np.zeros(3) if origin is None else np.array(origin)
    rotation_matrix = rotation.get_rotation_matrix()
    expected = np.array(
        [
            rotation_matrix @ (p - origin_for_reference) + origin_for_reference
            for p in points
        ]
    )
    assert_results_close(spline.control_points, expected)


def test_beamme_mesh_creation_functions_nurbs_utils_ensure_3d_splinepy_object_already_3d():
    """Ensure that a 3D splinepy object is left unchanged."""
    original_points = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    spline = MockSpline(original_points)
    spline_cp_initial = spline.control_points
    ensure_3d_splinepy_object(spline)
    # Control point object should be same identical object
    assert spline.control_points is spline_cp_initial


def test_beamme_mesh_creation_functions_nurbs_utils_ensure_3d_splinepy_object_from_1d(
    assert_results_close,
):
    """Ensure that a 1D splinepy object is converted to 3D by adding y=0 and
    z=0."""
    original_points = np.array([[1.0], [3.0]])
    spline = MockSpline(original_points)
    ensure_3d_splinepy_object(spline)
    expected = np.array([[1.0, 0.0, 0.0], [3.0, 0.0, 0.0]])
    assert_results_close(spline.control_points, expected)


def test_beamme_mesh_creation_functions_nurbs_utils_ensure_3d_splinepy_object_from_2d(
    assert_results_close,
):
    """Ensure that a 2D splinepy object is converted to 3D by adding z=0."""
    original_points = np.array([[1.0, 2.0], [3.0, 4.0]])
    spline = MockSpline(original_points)
    ensure_3d_splinepy_object(spline)
    expected = np.array([[1.0, 2.0, 0.0], [3.0, 4.0, 0.0]])
    assert_results_close(spline.control_points, expected)


def test_beamme_mesh_creation_functions_nurbs_utils_ensure_3d_splinepy_object_invalid_dimension():
    """Ensure that an invalid spline dimension raises ValueError."""
    spline = MockSpline(np.zeros((2, 4)))  # 4D control points
    with pytest.raises(ValueError, match="Splinepy object must be 1D, 2D or 3D"):
        ensure_3d_splinepy_object(spline)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test README.md."""

import os
import re

import pytest

from beamme.core.mesh import Mesh


@pytest.fixture(scope="session")
def extract_code_snippets(pytestconfig) -> tuple[dict[str, str], list[str]]:
    """Parse README.md for fenced code blocks marked as tests.

    Code fences may be written in two forms:
        ```python test
        # unnamed test block
        ...
        ```
    or
        ```python test:my_name
        # named test block
        ...
        ```

    Returns:
        A pair consisting of:
            - **snippets_named**: dict mapping snippet name (the part after
              `test:`) to the code string.
            - **snippets_unnamed**: list of code strings from unnamed test blocks.
    """
    readme_content = (pytestconfig.rootpath / "README.md").read_text(encoding="utf-8")

    # Match ```python test or ```python test:name
    pattern = re.compile(r"```python\s+test(?::([^\n]+))?\s*\r?\n(.*?)```", re.DOTALL)
    snippets_named = {}
    snippets_unnamed = []
    for i, (name, code) in enumerate(pattern.findall(readme_content)):
        code = code.strip()
        if name:
            snippets_named[name.strip()] = code
        else:
            snippets_unnamed.append(code)
    return snippets_named, snippets_unnamed


def test_readme_auto(extract_code_snippets):
    """Run all unnamed code snippets from README automatically."""
    _, snippets_unnamed = extract_code_snippets
    for code in snippets_unnamed:
        exec(code, {})


def test_other_readme_getting_started(
    extract_code_snippets,
    get_corresponding_reference_file_path,
    assert_results_close,
    tmp_path,
):
    """Test the getting started example in the README.md."""
    snippets_named, _ = extract_code_snippets
    os.chdir(tmp_path)
    globals = {}
    exec(snippets_named["getting_started"], globals)

    # The example creates an object `mesh` - check if it exists and is of type `Mesh`.
    assert "mesh" in globals
    assert isinstance(globals["mesh"], Mesh)

    # What we can do, is to check the created vtk output.
    ref_file = get_corresponding_reference_file_path(
        additional_identifier="beam", extension="vtu"
    )
    vtk_file = tmp_path / "getting_started_beam.vtu"
    assert_results_close(ref_file, vtk_file)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# cython: language_level=3
"""
This script contains functions that calculate points that are close together.
It uses cython to gain performance.
"""


import numpy as np
from libc.math cimport sqrt

cimport cython
cimport numpy as np

@cython.boundscheck(False) # Deactivate checking bounds of arrays.
@cython.wraparound(False)  # Deactivate negative indexing.
def find_close_points(np.ndarray[double, ndim=2] coords, double eps):
    """
    Finds coordinates that are within a tolerance of each other.

    Args
    ----
    coords: numpy array
        Array with the coordinates of the nodes.
    eps: float
        Tolerance to look for neighbors.

    Return
    ----
    has_partner: numpy array
        An array with integers, marking the set a node is part of. -1 means the
        node does not have a partner.
    partner: int
        Number of clusters found.
    """

    # Define types of variables for this function.
    cdef np.int64_t n_nodes, partner, this_is_partner, i, j, k, n_dim
    cdef double distance

    # Number of nodes and dimension of coordinates.
    n_nodes = np.shape(coords)[0]
    n_dim = np.shape(coords)[1]

    # This vector is -1 if a node does not belong to a pair, otherwise it is
    # the number of the pair.
    cdef np.ndarray[np.int64_t, ndim=1] has_partner = (
        np.zeros(n_nodes, dtype=np.int64) - 1)

    # Loop over nodes.
    partner = 0
    for i in range(n_nodes):
        this_is_partner = 0
        if has_partner[i] == -1:
            for j in range(i + 1, n_nodes):
                # Calculate the distance between the two nodes.
                distance = 0.
                for k in range(n_dim):
                    distance += (coords[i, k] - coords[j, k])**2
                distance = sqrt(distance)
                # Check if the distance is smaller than the threshold, and add
                # to has_partner list.
                if distance < eps:
                    this_is_partner = 1
                    if not has_partner[j] == -1:
                        raise RuntimeError('The case where a node connects two'
                            ' other nodes that are more than eps apart is not'
                            ' yet implemented. Check if the value for eps'
                            ' makes physical sense!')
                    has_partner[j] = partner
            # If this one has a partner set this node too.
            if this_is_partner == 1:
                has_partner[i] = partner
                partner += 1

    return has_partner, partner



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test the functionality of the core mesh."""

import numpy as np
import pytest

from beamme.core.conf import bme
from beamme.core.coupling import Coupling
from beamme.core.element_beam import Beam, Beam3
from beamme.core.geometry_set import GeometrySet
from beamme.core.mesh import Mesh
from beamme.core.node import Node
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line


def test_beamme_core_mesh_get_nodes_by_function(
    get_default_test_beam_material, assert_results_close
):
    """Check if the get_nodes_by_function method of Mesh works properly."""

    def get_nodes_at_x(node, x_value):
        """True for all coordinates at a certain x value."""
        if np.abs(node.coordinates[0] - x_value) < 1e-10:
            return True
        else:
            return False

    mat = get_default_test_beam_material(material_type="base")

    mesh = Mesh()
    create_beam_mesh_line(mesh, Beam3, mat, [0, 0, 0], [5, 0, 0], n_el=5)
    create_beam_mesh_line(mesh, Beam3, mat, [0, 1, 0], [10, 1, 0], n_el=10)

    nodes = mesh.get_nodes_by_function(get_nodes_at_x, 1.0)
    assert 2 == len(nodes)
    for node in nodes:
        assert_results_close(1.0, node.coordinates[0])


def test_beamme_core_mesh_add_checks():
    """This test checks that Mesh raises an error when double objects are added
    to the mesh."""

    # Mesh instance for this test.
    mesh = Mesh()

    # Create basic objects that will be added to the mesh.
    node = Node([0, 1.0, 2.0])
    element = Beam()
    mesh.add(node)
    mesh.add(element)

    # Create objects based on basic mesh items.
    coupling = Coupling(mesh.nodes, bme.bc.point_coupling, bme.coupling_dof.fix)
    coupling_penalty = Coupling(
        mesh.nodes, bme.bc.point_coupling_penalty, bme.coupling_dof.fix
    )
    geometry_set = GeometrySet(mesh.elements)
    mesh.add(coupling)
    mesh.add(coupling_penalty)
    mesh.add(geometry_set)

    # Add the objects again and check for errors.
    with pytest.raises(ValueError, match="The node is already in this mesh!"):
        mesh.add(node)
    with pytest.raises(ValueError, match="The element is already in this mesh!"):
        mesh.add(element)
    with pytest.raises(ValueError, match="The item is already in this container!"):
        mesh.add(coupling)
    with pytest.raises(ValueError, match="The item is already in this container!"):
        mesh.add(coupling_penalty)
    with pytest.raises(ValueError, match="The item is already in this container!"):
        mesh.add(geometry_set)


def test_beamme_core_mesh_multiple_couple_nodes(get_default_test_beam_material):
    """The current implementation can handle more than one coupling on a node
    correctly, therefore we check this here."""

    # Create mesh object
    mesh = Mesh()
    mat = get_default_test_beam_material(material_type="reissner")
    mesh.add(mat)

    # Add two beams to create an elbow structure. The beams each have a
    # node at the intersection
    beam_set_1 = create_beam_mesh_line(mesh, Beam3, mat, [0, 0, 0], [1, 0, 0])
    beam_set_2 = create_beam_mesh_line(mesh, Beam3, mat, [1, 0, 0], [1, 1, 0])

    # Call coupling twice -> this will create two coupling objects for the
    # corner node
    mesh.couple_nodes()
    mesh.couple_nodes()

    # Check that we have two point couplings and that they both couple the end and
    # start node of the created lines.
    couplings = mesh.boundary_conditions[bme.bc.point_coupling, bme.geo.point]
    assert len(couplings) == 2
    for coupling in couplings:
        coupling_nodes = coupling.geometry_set.get_points()
        assert len(coupling_nodes) == 2
        assert beam_set_1["end"].get_points()[0] in coupling_nodes
        assert beam_set_2["start"].get_points()[0] in coupling_nodes



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module defines functions that can be used to add header information to
an input file."""

from typing import Any as _Any
from typing import List as _List
from typing import Union as _Union

from beamme.core.conf import bme as _bme
from beamme.four_c.input_file import InputFile as _InputFile


def _get_segmentation_strategy(segmentation):
    """Get the 4C string for a geometry pair strategy."""
    if segmentation:
        return "segmentation"
    else:
        return "gauss_point_projection_without_boundary_segmentation"


def set_runtime_output(
    input_file,
    *,
    output_solid=True,
    output_stress_strain=False,
    btsvmt_output=True,
    btss_output=True,
    output_triad=True,
    every_iteration=False,
    absolute_beam_positions=True,
    element_owner=True,
    element_gid=True,
    element_mat_id=True,
    output_energy=False,
    output_strains=True,
):
    """Set the basic runtime output options.

    Args
    ----
    input_file:
        Input file that the options will be added to.
    output_solid: bool
        If the solid output should be written at runtime.
    output_stress_strain: bool
        If stress and strain output should be written for the solid.
    btsvmt_output: bool
        If the output for btsvmt should be written.
    btss_output: bool
        If the output for beam-to-surface coupling should be written.
    output_triad: bool
        If the triads along the beam should be written.
    every_iteration: int
        If output at every Newton iteration should be written.
    absolute_beam_positions: bool
        If the beams should be written at the current position or always at
        the reference position.
    element_owner: bool
        If the owing rank of each element should be output (currently
        only affects the solid elements in 4C, beam element owners are
        written by default).
    element_gid: bool
        If the 4C internal GID of each element should be output.
    element_mat_id: bool
        If the 4C internal material ID of each element should be output.
    output_energy: bool
        If the energy output from 4C should be activated.
    output_strains: bool
        If the strains in the Gauss points should be output.
    """

    # Set the basic runtime output options.
    input_file.add(
        {
            "IO/RUNTIME VTK OUTPUT": {
                "OUTPUT_DATA_FORMAT": "binary",
                "INTERVAL_STEPS": 1,
                "EVERY_ITERATION": every_iteration,
            }
        }
    )

    # Set the structure runtime output options
    input_file.add(
        {
            "IO/RUNTIME VTK OUTPUT/STRUCTURE": {
                "OUTPUT_STRUCTURE": output_solid,
                "DISPLACEMENT": True,
                "STRESS_STRAIN": output_stress_strain,
                "ELEMENT_OWNER": element_owner,
                "ELEMENT_GID": element_gid,
                "ELEMENT_MAT_ID": element_mat_id,
            }
        }
    )

    # Set the beam runtime output options
    input_file.add(
        {
            "IO/RUNTIME VTK OUTPUT/BEAMS": {
                "OUTPUT_BEAMS": True,
                "DISPLACEMENT": True,
                "USE_ABSOLUTE_POSITIONS": absolute_beam_positions,
                "TRIAD_VISUALIZATIONPOINT": output_triad,
                "STRAINS_GAUSSPOINT": output_strains,
                "ELEMENT_GID": element_gid,
            }
        }
    )

    if btsvmt_output:
        # Set the beam to solid volume mesh tying runtime output options.
        input_file.add(
            {
                "BEAM INTERACTION/BEAM TO SOLID VOLUME MESHTYING/RUNTIME VTK OUTPUT": {
                    "WRITE_OUTPUT": True,
                    "NODAL_FORCES": True,
                    "MORTAR_LAMBDA_DISCRET": True,
                    "MORTAR_LAMBDA_CONTINUOUS": True,
                    "MORTAR_LAMBDA_CONTINUOUS_SEGMENTS": 5,
                    "SEGMENTATION": True,
                    "INTEGRATION_POINTS": True,
                }
            }
        )

    if btss_output:
        # Set the beam to solid surface coupling runtime output options.
        input_file.add(
            {
                "BEAM INTERACTION/BEAM TO SOLID SURFACE/RUNTIME VTK OUTPUT": {
                    "WRITE_OUTPUT": True,
                    "NODAL_FORCES": True,
                    "MORTAR_LAMBDA_DISCRET": True,
                    "MORTAR_LAMBDA_CONTINUOUS": True,
                    "MORTAR_LAMBDA_CONTINUOUS_SEGMENTS": 5,
                    "SEGMENTATION": True,
                    "INTEGRATION_POINTS": True,
                    "AVERAGED_NORMALS": True,
                }
            }
        )

    if output_energy:
        input_file["STRUCTURAL DYNAMIC"]["RESEVERYERGY"] = 1


def set_beam_to_solid_meshtying(
    input_file,
    interaction_type,
    *,
    contact_discretization=None,
    segmentation=True,
    segmentation_search_points=2,
    couple_restart=False,
    mortar_shape="none",
    n_gauss_points=6,
    n_integration_points_circ=None,
    penalty_parameter=None,
    coupling_type=None,
    binning_parameters: dict = {},
):
    """Set the beam to solid meshtying options.

    Args
    ----
    input_file:
        Input file that the options will be added to.
    interaction_type: BoundaryCondition
        Type of beam-to-solid interaction.
    contact_discretization: str
        Type of contact (mortar, Gauss point, ...)
    segmentation: bool
        If segmentation should be used in the numerical integration.
    segmentation_search_points: int
        Number of search points for segmentation.
    couple_restart: bool
        If the restart configuration should be used for the coupling
    mortar_shape: str
        Type of shape function for mortar discretization.
    n_gauss_points: int
        Number of Gauss points for numerical integration.
    n_integration_points_circ: int
        Number of integration points along the circumference of the cross
        section.
    penalty_parameter: float
        Penalty parameter for contact enforcement.
    coupling_type: str
        Type of coupling for beam-to-surface coupling.
    binning_parameters:
        Keyword parameters for the binning section
    """

    # Set the beam contact options.
    # check if these keys are already set, otherwise set them
    if (
        "BEAM INTERACTION" not in input_file
        or input_file["BEAM INTERACTION"].get("REPARTITIONSTRATEGY") != "everydt"
    ):
        input_file.add({"BEAM INTERACTION": {"REPARTITIONSTRATEGY": "everydt"}})

    set_binning_strategy_section(
        input_file,
        **binning_parameters,
    )

    # Add the beam to solid volume mesh tying options.
    bts_parameters = {}
    if interaction_type == _bme.bc.beam_to_solid_volume_meshtying:
        bts_section_name = "BEAM INTERACTION/BEAM TO SOLID VOLUME MESHTYING"
    elif interaction_type == _bme.bc.beam_to_solid_surface_meshtying:
        bts_section_name = "BEAM INTERACTION/BEAM TO SOLID SURFACE MESHTYING"
        if coupling_type is not None:
            bts_parameters["COUPLING_TYPE"] = coupling_type
    else:
        raise ValueError(
            "Got wrong beam-to-solid mesh tying type. "
            f"Got {interaction_type} of type {type(interaction_type)}."
        )
    bts_parameters["CONSTRAINT_STRATEGY"] = "penalty"
    if penalty_parameter is not None:
        bts_parameters["PENALTY_PARAMETER"] = penalty_parameter
    bts_parameters["GAUSS_POINTS"] = n_gauss_points

    if contact_discretization == "mortar":
        bts_parameters["CONTACT_DISCRETIZATION"] = "mortar"
        bts_parameters["MORTAR_SHAPE_FUNCTION"] = mortar_shape
        segmentation_strategy = _get_segmentation_strategy(segmentation)
    elif contact_discretization == "gp":
        bts_parameters["CONTACT_DISCRETIZATION"] = "gauss_point_to_segment"
        segmentation_strategy = _get_segmentation_strategy(segmentation)
    elif contact_discretization == "circ":
        bts_parameters["CONTACT_DISCRETIZATION"] = "gauss_point_cross_section"
        bts_parameters["INTEGRATION_POINTS_CIRCUMFERENCE"] = n_integration_points_circ
        segmentation_strategy = "gauss_point_projection_cross_section"
    else:
        raise ValueError(
            f'Wrong contact_discretization "{contact_discretization}" given!'
        )

    bts_parameters["GEOMETRY_PAIR_STRATEGY"] = segmentation_strategy
    bts_parameters["GEOMETRY_PAIR_SEGMENTATION_SEARCH_POINTS"] = (
        segmentation_search_points
    )
    if interaction_type == _bme.bc.beam_to_solid_volume_meshtying:
        bts_parameters["COUPLE_RESTART_STATE"] = couple_restart

    input_file.add({bts_section_name: bts_parameters})


def set_header_static(
    input_file: _InputFile,
    *,
    time_step: float | None = None,
    n_steps: int | None = None,
    total_time: float | None = None,
    max_iter: int = 20,
    tol_residuum: float = 1e-8,
    tol_increment: float = 1e-10,
    restart_every: int = 1,
    load_lin: bool = False,
    write_bin: bool = False,
    write_stress: str = "no",
    write_strain: str = "no",
    predictor: str = "TangDis",
    prestress: str = "None",
    prestress_time: float = 0,
    create_nox_file: bool = True,
):
    """Set the default parameters for a static structure analysis.

    At least two of the three time stepping keyword arguments ["time_step",
    "n_steps", "total_time"] have to be set.

    Args:
        input_file:
            Input file that the options will be added to.
        time_step:
            Time increment per step.
        n_steps:
            Number of time steps.
        total_time:
            Total time of simulation
        max_iter:
            Maximal number of Newton iterations.
        tol_residuum:
            Tolerance for the convergence of the residuum.
        tol_increment:
            Tolerance for the convergence of the displacement increment.
        load_lin:
            If the load_lin option should be set.
        write_bin:
            If binary output should be written.
        restart_every:
            Frequency for writing restart output.
        write_stress:
            If and which stress output to write
        write_strain:
            If and which strain output to write
        predictor:
            Type of predictor to be used
        prestress:
            Type of prestressing strategy to be used
        prestress_time:
            Prestress Time
        create_nox_file:
            If the nonlinear solver parameters should be set via a NOX xml file or
            directly in the input file.
    """

    input_file_parameters: dict[str, _Any] = {}

    # Set the parameters for a static analysis.
    input_file_parameters["PROBLEM TYPE"] = {"PROBLEMTYPE": "Structure"}
    input_file_parameters["IO"] = {
        "OUTPUT_BIN": write_bin,
        "STRUCT_DISP": False,
        "STRUCT_STRESS": write_stress,
        "STRUCT_STRAIN": write_strain,
        "VERBOSITY": "Standard",
    }

    # Set the time step parameters
    given_time_arguments = sum(
        1 for arg in (time_step, n_steps, total_time) if arg is not None
    )
    if given_time_arguments < 2:
        raise ValueError(
            'At least two of the following arguments "time_step", "n_steps" or '
            '"total_time" are required'
        )
    elif time_step is None and total_time is not None and n_steps is not None:
        time_step = total_time / n_steps
    elif n_steps is None and total_time is not None and time_step is not None:
        n_steps = round(total_time / time_step)
    elif total_time is None and time_step is not None and n_steps is not None:
        total_time = time_step * n_steps

    input_file_parameters["STRUCTURAL DYNAMIC"] = {
        "LINEAR_SOLVER": 1,
        "INT_STRATEGY": "Standard",
        "DYNAMICTYPE": "Statics",
        "PREDICT": predictor,
        "PRESTRESS": prestress,
        "PRESTRESSTIME": prestress_time,
        "TIMESTEP": time_step,
        "NUMSTEP": n_steps,
        "MAXTIME": total_time,
        "LOADLIN": load_lin,
        "RESTARTEVERY": restart_every,
    }
    input_file_parameters["SOLVER 1"] = {
        "NAME": "Structure_Solver",
        "SOLVER": "Superlu",
    }

    # Set the solver parameters.
    if create_nox_file:
        # Set the contents of the NOX xml file.
        nox_xml_contents = f"""
            <ParameterList name="Status Test">
            <!-- Outer Status Test: This test is an OR combination of the structural convergence and the maximum number of iterations -->
            <ParameterList name="Outer Status Test">
            <Parameter name="Test Type"       type="string" value="Combo"/>
            <Parameter name="Combo Type"      type="string" value="OR" />
            <!-- Structural convergence is an AND combination of the residuum and step update -->
            <ParameterList name="Test 0">
                <Parameter name="Test Type" type="string" value="Combo" />
                <Parameter name="Combo Type" type="string" value="AND" />
                <!-- BEGIN: Combo AND - Test 0: "NormF" -->
                <ParameterList name="Test 0">
                    <Parameter name="Test Type"  type="string" value="NormF" />
                    <!-- NormF - Quantity 0: Check the right-hand-side norm of the structural quantities -->
                    <ParameterList name="Quantity 0">
                    <Parameter name="Quantity Type"  type="string" value="Structure" />
                    <Parameter name="Tolerance Type" type="string" value="Absolute" />
                    <Parameter name="Tolerance"      type="double" value="{tol_residuum}" />
                    <Parameter name="Norm Type"      type="string" value="Two Norm" />
                    <Parameter name="Scale Type"     type="string" value="Scaled" />
                    </ParameterList>
                </ParameterList>
                <!-- END: Combo AND - Test 0: "NormF" -->
                <!-- BEGIN: Combo AND - Test 1: "NormWRMS" -->
                <ParameterList name="Test 1">
                    <Parameter name="Test Type"        type="string" value="NormUpdate" />
                    <!-- NormWRMS - Quantity 0: Check the increment of the structural displacements -->
                    <ParameterList name="Quantity 0">
                    <Parameter name="Quantity Type"  type="string" value="Structure" />
                    <Parameter name="Tolerance Type" type="string" value="Absolute" />
                    <Parameter name="Tolerance"      type="double" value="{tol_increment}" />
                    <Parameter name="Norm Type"      type="string" value="Two Norm" />
                    <Parameter name="Scale Type"     type="string" value="Scaled" />
                    </ParameterList>
                </ParameterList>
                <!-- END: Combo AND - Test 1: "NormWRMS" -->
                </ParameterList>
                <!-- END: Combo 0 - Test 0: "Combo" -->
            <!-- BEGIN: Combo OR - Test 1: "MaxIters" -->
            <ParameterList name="Test 1">
                <Parameter name="Test Type"          type="string" value="MaxIters" />
                <Parameter name="Maximum Iterations" type="int"    value="{max_iter}" />
            </ParameterList> <!--END: "MaxIters" -->
            </ParameterList>
            </ParameterList>
            """

        input_file_parameters["STRUCT NOX/Printing"] = {
            "Error": True,
            "Inner Iteration": False,
            "Details": True,
            "Linear Solver Details": True,
            "Test Details": True,
        }

        # Set the xml content in the input file.
        input_file.nox_xml_contents = nox_xml_contents

    else:
        input_file_parameters["STRUCTURAL DYNAMIC"]["MAXITER"] = max_iter
        input_file_parameters["STRUCTURAL DYNAMIC"]["TOLRES"] = tol_residuum
        input_file_parameters["STRUCTURAL DYNAMIC"]["TOLDISP"] = tol_increment

    input_file.add(input_file_parameters)


def set_binning_strategy_section(
    input_file: _InputFile,
    binning_bounding_box: _Union[_List[int], None] = None,
    binning_cutoff_radius: _Union[float, None] = None,
):
    """Set binning strategy in section of the input file.

    Args
    ----
    input_file:
        Input file that the options will be added to.
    binning_bounding_box:
        List with the limits of the bounding box.
    binning_cutoff_radius:
        Maximal influence radius of pair elements.
    """

    if binning_bounding_box is not None and binning_cutoff_radius is not None:
        binning_bounding_box_string = " ".join(
            [str(val) for val in binning_bounding_box]
        )

        input_file.add(
            {
                "BINNING STRATEGY": {
                    "BIN_SIZE_LOWER_BOUND": binning_cutoff_radius,
                    "DOMAINBOUNDINGBOX": binning_bounding_box_string,
                }
            }
        )
    elif [binning_bounding_box, binning_cutoff_radius].count(None) == 2:
        return
    else:
        raise ValueError(
            f"The variables binning_bounding_box {binning_bounding_box} and binning_cutoff_radius {binning_cutoff_radius} must both be set."
        )


def set_beam_interaction_section(
    input_file: _InputFile,
    *,
    repartition_strategy: str = "everydt",
    search_strategy: str = "bounding_volume_hierarchy",
):
    """Set beam interaction section in input file.

    Args
    ----
    input_file:
        Input file that the options will be added to.
    repartition_strategy:
        Type of employed repartitioning strategy
        Options: "adaptive" or "everydt"
    search_strategy:
        Type of search strategy used for finding coupling pairs.
        Options: "bruteforce_with_binning", "bounding_volume_hierarchy"
    """

    input_file.add(
        {
            "BEAM INTERACTION": {
                "REPARTITIONSTRATEGY": repartition_strategy,
                "SEARCH_STRATEGY": search_strategy,
            }
        }
    )


def set_beam_contact_runtime_output(
    input_file: _InputFile, *, every_iteration: bool = False
):
    """Output the beam-to-beam contact forces and gaps with runtime output.

    input_file:
        Input file that the options will be added to.
    every_iteration:
        If output at every Newton iteration should be written.
    """

    input_file.add(
        {
            "BEAM INTERACTION/BEAM TO BEAM CONTACT/RUNTIME VTK OUTPUT": {
                "VTK_OUTPUT_BEAM_CONTACT": True,
                "EVERY_ITERATION": every_iteration,
                "INTERVAL_STEPS": 1,
                "CONTACT_FORCES": True,
                "GAPS": True,
            }
        }
    )


def set_beam_contact_section(
    input_file: _InputFile,
    *,
    interaction_strategy: str = "penalty",
    btb_penalty: float = 0,
    btb_line_penalty: float = 0,
    per_shift_angle: list[float] = [70, 80],
    par_shift_angle: list[float] = [70, 80],
    b_seg_angle: float = 12,
    num_integration: int = 5,
    penalty_law: str = "LinPosQuadPen",
    penalty_regularization_g0: float = 0,
    penalty_regularization_f0: float = 0,
    penalty_regularization_c0: float = 0,
    binning_parameters: dict = {},
    beam_interaction_parameters: dict = {},
):
    """Set default beam contact section, for more and updated details see
    respective input file within 4C. Parameters for set_binning_strategy and
    set_beam_interaction may be forwarded as keyword arguments.

    Args
    ----
    input_file:
        Input file that the options will be added to.
    interaction_strategy:
        Type of employed solving strategy
        Options: "none", "penalty" or "gmshonly"
    btb_penalty: double
        Penalty parameter for beam-to-beam point contact
    btb_line_penalty:
        Penalty parameter per unit length for beam-to-beam line contact
    per_shift_angle:
        Lower and upper shift angle (in degrees) for penalty scaling of large-angle-contact
    par_shift_angle:
        Lower and upper shift angle (in degrees) for penalty scaling of small-angle-contact
    b_seg_angle:
        Maximal angle deviation allowed for contact search segmentation
    num_integration:
        Number of integration intervals per element
    penalty_law:
        Penalty Law Options: "LinPen", "QuadPen", "LinNegQuadPen", "LinPosQuadPen", "LinPosCubPen", "LinPosDoubleQuadPen", "LinPosExpPen"
    penalty_regularization_g0:
        First penalty regularization parameter G0
    penalty_regularization_f0:
        Second penalty regularization parameter F0
    penalty_regularization_c0:
        Third penalty regularization parameter C0
    binning_parameters:
        Keyword parameters for the binning section
    beam_interaction_parameters:
        Keyword parameters for the beam-contact section
    """

    if len(per_shift_angle) != 2:
        raise ValueError(
            "Please provide lower and upper value of BEAMS_PERPSHIFTANGLE."
        )

    if len(par_shift_angle) != 2:
        raise ValueError("Please provide lower and upper value of BEAMS_PARSHIFTANGLE.")

    input_file.add(
        {
            "BEAM INTERACTION/BEAM TO BEAM CONTACT": {
                "BEAMS_BTBPENALTYPARAM": btb_penalty,
                "BEAMS_BTBLINEPENALTYPARAM": btb_line_penalty,
                "BEAMS_SEGCON": True,
                "BEAMS_PERPSHIFTANGLE1": per_shift_angle[0],
                "BEAMS_PERPSHIFTANGLE2": per_shift_angle[1],
                "BEAMS_PARSHIFTANGLE1": par_shift_angle[0],
                "BEAMS_PARSHIFTANGLE2": par_shift_angle[1],
                "BEAMS_SEGANGLE": b_seg_angle,
                "BEAMS_NUMINTEGRATIONINTERVAL": num_integration,
                "BEAMS_PENALTYLAW": penalty_law,
                "BEAMS_PENREGPARAM_G0": penalty_regularization_g0,
                "BEAMS_PENREGPARAM_F0": penalty_regularization_f0,
                "BEAMS_PENREGPARAM_C0": penalty_regularization_c0,
            }
        }
    )

    # beam contact needs a binning strategy
    set_binning_strategy_section(input_file, **binning_parameters)

    # beam contact needs interaction strategy
    set_beam_interaction_section(input_file, **beam_interaction_parameters)


def add_result_description(
    input_file: _InputFile,
    displacements: _List,
    node_ids: _List[int],
    *,
    tol: float = 1e-10,
):
    """Add result descriptions for structure problems to the input file.

    Args:
        input_file: Input file to add the result description to
        displacements: Array with the displacements (n_nodes x 3)
        node_ids: List with the IDs of the nodes to check
        tol: Tolerance
    """
    result_descriptions = []

    for i_node, node in enumerate(node_ids):
        for i_dir, direction in enumerate(["x", "y", "z"]):
            result_descriptions.append(
                {
                    "STRUCTURE": {
                        "DIS": "structure",
                        "NODE": node,
                        "QUANTITY": f"disp{direction}",
                        "VALUE": displacements[i_node][i_dir],
                        "TOLERANCE": tol,
                    },
                }
            )

    input_file.add({"RESULT DESCRIPTION": result_descriptions})



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test general functionality of the core geometry set
class with end-to-end integration tests."""

import pytest

from beamme.core.boundary_condition import BoundaryCondition
from beamme.core.conf import bme
from beamme.core.geometry_set import GeometrySet, GeometrySetNodes
from beamme.core.mesh import Mesh
from beamme.core.rotation import Rotation
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line


@pytest.mark.parametrize("use_nodal_geometry_sets", [True, False])
def test_integration_core_geometry_set_replace_nodes_geometry_set(
    get_bc_data,
    get_default_test_beam_material,
    use_nodal_geometry_sets,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test case for coupling of nodes, and reusing the identical nodes."""

    bme.check_overlapping_elements = False

    mat = get_default_test_beam_material(material_type="reissner")
    rot = Rotation([1, 2, 43], 213123)

    # Create a beam with two elements. Once immediately and once as two
    # beams with couplings.
    mesh_ref = Mesh()
    mesh_couple = Mesh()

    # Create a simple beam.
    create_beam_mesh_line(mesh_ref, Beam3rHerm2Line3, mat, [0, 0, 0], [2, 0, 0], n_el=2)
    create_beam_mesh_line(mesh_couple, Beam3rHerm2Line3, mat, [0, 0, 0], [1, 0, 0])
    create_beam_mesh_line(mesh_couple, Beam3rHerm2Line3, mat, [1, 0, 0], [2, 0, 0])

    ref_nodes = list(mesh_ref.nodes)
    coupling_nodes = list(mesh_couple.nodes)

    # Add a set with all nodes, to check that the nodes in the
    # boundary condition are replaced correctly.
    if use_nodal_geometry_sets:
        mesh_ref.add(GeometrySetNodes(bme.geo.line, ref_nodes))
        mesh_ref.add(GeometrySetNodes(bme.geo.point, ref_nodes))
        mesh_couple.add(GeometrySetNodes(bme.geo.line, coupling_nodes))
        mesh_couple.add(GeometrySetNodes(bme.geo.point, coupling_nodes))
    else:
        mesh_ref.add(GeometrySet(mesh_ref.elements))
        mesh_ref.add(GeometrySet(ref_nodes))
        mesh_couple.add(GeometrySet(mesh_couple.elements))
        mesh_couple.add(GeometrySet(coupling_nodes))

    # Add another set with all nodes, this time only the coupling node
    # that will be kept is in this set.
    coupling_nodes_without_replace_node = list(coupling_nodes)
    del coupling_nodes_without_replace_node[3]
    if use_nodal_geometry_sets:
        mesh_ref.add(GeometrySetNodes(bme.geo.point, ref_nodes))
        mesh_couple.add(
            GeometrySetNodes(bme.geo.point, coupling_nodes_without_replace_node)
        )
    else:
        mesh_ref.add(GeometrySet(ref_nodes))
        mesh_couple.add(GeometrySet(coupling_nodes_without_replace_node))

    # Add another set with all nodes, this time only the coupling node
    # that will be replaced is in this set.
    coupling_nodes_without_replace_node = list(coupling_nodes)
    del coupling_nodes_without_replace_node[2]
    if use_nodal_geometry_sets:
        mesh_ref.add(GeometrySetNodes(bme.geo.point, ref_nodes))
        mesh_couple.add(
            GeometrySetNodes(bme.geo.point, coupling_nodes_without_replace_node)
        )
    else:
        mesh_ref.add(GeometrySet(ref_nodes))
        mesh_couple.add(GeometrySet(coupling_nodes_without_replace_node))

    # Rotate both meshes
    mesh_ref.rotate(rot)
    mesh_couple.rotate(rot)

    # Couple the coupling mesh.
    mesh_couple.couple_nodes(
        coupling_dof_type=bme.coupling_dof.fix, reuse_matching_nodes=True
    )

    # Compare the meshes.
    assert_results_close(mesh_ref, mesh_couple)
    assert_results_close(
        get_corresponding_reference_file_path(additional_identifier="1"), mesh_couple
    )

    # Create two overlapping beams. This is to test that the middle nodes
    # are not coupled.
    mesh_ref = Mesh()
    mesh_couple = Mesh()

    # Create a simple beam.
    set_ref = create_beam_mesh_line(
        mesh_ref, Beam3rHerm2Line3, mat, [0, 0, 0], [1, 0, 0]
    )
    create_beam_mesh_line(
        mesh_ref,
        Beam3rHerm2Line3,
        mat,
        [0, 0, 0],
        [1, 0, 0],
        start_node=set_ref["start"],
        end_node=set_ref["end"],
    )
    create_beam_mesh_line(mesh_couple, Beam3rHerm2Line3, mat, [0, 0, 0], [1, 0, 0])
    create_beam_mesh_line(mesh_couple, Beam3rHerm2Line3, mat, [0, 0, 0], [1, 0, 0])

    # Rotate both meshes
    mesh_ref.rotate(rot)
    mesh_couple.rotate(rot)

    # Couple the coupling mesh.
    mesh_couple.couple_nodes(
        coupling_dof_type=bme.coupling_dof.fix, reuse_matching_nodes=True
    )

    # Compare the meshes.
    assert_results_close(mesh_ref, mesh_couple)
    assert_results_close(
        get_corresponding_reference_file_path(additional_identifier="2"), mesh_couple
    )

    # Create a beam with two elements. Once immediately and once as two
    # beams with couplings.
    mesh_ref = Mesh()
    mesh_couple = Mesh()

    # Create a simple beam.
    create_beam_mesh_line(mesh_ref, Beam3rHerm2Line3, mat, [0, 0, 0], [2, 0, 0], n_el=2)
    create_beam_mesh_line(mesh_couple, Beam3rHerm2Line3, mat, [0, 0, 0], [1, 0, 0])
    create_beam_mesh_line(mesh_couple, Beam3rHerm2Line3, mat, [1, 0, 0], [2, 0, 0])

    # Create set with all the beam nodes.
    if use_nodal_geometry_sets:
        node_set_1_ref = GeometrySetNodes(bme.geo.line, mesh_ref.nodes)
        node_set_2_ref = GeometrySetNodes(bme.geo.line, mesh_ref.nodes)
        node_set_1_couple = GeometrySetNodes(bme.geo.line, mesh_couple.nodes)
        node_set_2_couple = GeometrySetNodes(bme.geo.line, mesh_couple.nodes)
    else:
        node_set_1_ref = GeometrySet(mesh_ref.elements)
        node_set_2_ref = GeometrySet(mesh_ref.elements)
        node_set_1_couple = GeometrySet(mesh_couple.elements)
        node_set_2_couple = GeometrySet(mesh_couple.elements)

    # Create connecting beams.
    create_beam_mesh_line(mesh_ref, Beam3rHerm2Line3, mat, [1, 0, 0], [2, 2, 2])
    create_beam_mesh_line(mesh_ref, Beam3rHerm2Line3, mat, [1, 0, 0], [2, -2, -2])
    create_beam_mesh_line(mesh_couple, Beam3rHerm2Line3, mat, [1, 0, 0], [2, 2, 2])
    create_beam_mesh_line(mesh_couple, Beam3rHerm2Line3, mat, [1, 0, 0], [2, -2, -2])

    # Rotate both meshes
    mesh_ref.rotate(rot)
    mesh_couple.rotate(rot)

    # Couple the mesh.
    mesh_ref.couple_nodes(coupling_dof_type=bme.coupling_dof.fix)
    mesh_couple.couple_nodes(
        coupling_dof_type=bme.coupling_dof.fix, reuse_matching_nodes=True
    )

    # Add the node sets.
    mesh_ref.add(node_set_1_ref)
    mesh_couple.add(node_set_1_couple)

    # Add BCs.
    mesh_ref.add(
        BoundaryCondition(node_set_2_ref, get_bc_data(), bc_type=bme.bc.neumann)
    )
    mesh_couple.add(
        BoundaryCondition(node_set_2_couple, get_bc_data(), bc_type=bme.bc.neumann)
    )

    # Compare the meshes.
    assert_results_close(mesh_ref, mesh_couple)
    assert_results_close(
        get_corresponding_reference_file_path(additional_identifier="3"), mesh_couple
    )
